"""
Filename: constants.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Constants for OpenSpliceAI.
"""

CL_max=10000
# Maximum nucleotide context length (CL_max/2 on either side of the 
# position of interest)
# CL_max should be an even number

SL=5000
# Sequence length of SpliceAIs (SL+CL will be the input length and
# SL will be the output length)
#############################
# Global variable definition
############################## 
EPOCH_NUM = 10





__author__ = "Kuan-Hao Chao"
__contact__ = "kuanhao.chao@gmail.com"
__copyright__ = "Copyright 2025, Kuan-Hao Chao, Alan Mao"
__credits__ = ["Kuan-Hao Chao", "Alan Mao"]
__date__ = "2025/03/21"
__deprecated__ = False
__email__ =  "kuanhao.chao@gmail.com"
__license__ = "GPL-3.0"
__maintainer__ = "developer"
__status__ = "Kuan-Hao Chao"
__version__ = "0.0.5"


from importlib.resources import files
import pandas as pd
import numpy as np
from pyfaidx import Fasta
import logging
import platform
import os, glob
from openspliceai.train_base.openspliceai import SpliceAI
from openspliceai.constants import *
from openspliceai.predict.predict import *
from openspliceai.predict.utils import *
    
##############################################
## LOADING PYTORCH AND KERAS MODELS
##############################################

def setup_device():
        """Select computation device based on availability."""
        device_str = "cuda" if torch.cuda.is_available() else "mps" if platform.system() == "Darwin" else "cpu"
        return torch.device(device_str)

def load_pytorch_models(model_path, CL):
    """
    Loads a SpliceAI PyTorch model from given state, inferring device.
    
    Params:
    - model_path (str): Path to the model state dict, or a directory of models
    - CL (int): Context length parameter for model conversion.
    
    Returns:
    - loaded_models (list): SpliceAI model(s) loaded with given state.
    """
    
    def load_model(device, flanking_size):
        """Loads the given model."""
        # Hyper-parameters:
        # L: Number of convolution kernels
        # W: Convolution window size in each residual unit
        # AR: Atrous rate in each residual unit
        L = 32
        W = np.asarray([11, 11, 11, 11])
        AR = np.asarray([1, 1, 1, 1])
        N_GPUS = 2
        BATCH_SIZE = 18*N_GPUS

        if int(flanking_size) == 80:
            W = np.asarray([11, 11, 11, 11])
            AR = np.asarray([1, 1, 1, 1])
            BATCH_SIZE = 18*N_GPUS
        elif int(flanking_size) == 400:
            W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11])
            AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4])
            BATCH_SIZE = 18*N_GPUS
        elif int(flanking_size) == 2000:
            W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11,
                            21, 21, 21, 21])
            AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4,
                            10, 10, 10, 10])
            BATCH_SIZE = 12*N_GPUS
        elif int(flanking_size) == 10000:
            W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11,
                            21, 21, 21, 21, 41, 41, 41, 41])
            AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4,
                            10, 10, 10, 10, 25, 25, 25, 25])
            BATCH_SIZE = 6*N_GPUS

        CL = 2 * np.sum(AR*(W-1))

        print(f"\t[INFO] Context nucleotides {CL}")
        print(f"\t[INFO] Sequence length (output): {SL}")
        
        model = SpliceAI(L, W, AR).to(device)
        params = {'L': L, 'W': W, 'AR': AR, 'CL': CL, 'SL': SL, 'BATCH_SIZE': BATCH_SIZE, 'N_GPUS': N_GPUS}

        return model, params
    
    # Setup device
    device = setup_device()
    
    # Load all model state dicts given the supplied model path
    if os.path.isdir(model_path):
        model_files = glob.glob(os.path.join(model_path, '*.p[th]')) # gets all PyTorch models from supplied directory
        if not model_files:
            logging.error(f"No PyTorch model files found in directory: {model_path}")
            exit()
            
        models = []
        for model_file in model_files:
            try:
                model = torch.load(model_file, map_location=device)
                models.append(model)
            except Exception as e:
                logging.error(f"Error loading PyTorch model from file {model_file}: {e}. Skipping...")
                
        if not models:
            logging.error(f"No valid PyTorch models found in directory: {model_path}")
            exit()
    
    elif os.path.isfile(model_path):
        try:
            models = [torch.load(model_path, map_location=device)]
        except Exception as e:
            logging.error(f"Error loading PyTorch model from file {model_path}: {e}.")
            exit()
        
    else:
        logging.error(f"Invalid path: {model_path}")
        exit()
    
    # Load state of model to device
    # NOTE: supplied model paths should be state dicts, not model files  
    loaded_models = []
    
    mismatch_hint = (
        "[HINT] Ensure the checkpoint was trained with the same flanking size "
        f"({CL}) that you supplied. For packaged weights, choose the model from "
        f"the `{CL}nt` directory."
    )

    for state_dict in models:
        model, params = load_model(device, CL)  # loads new SpliceAI model with correct hyperparams
        try:
            model.load_state_dict(state_dict)   # loads state dict
        except RuntimeError as e:
            err_msg = str(e)
            if "size mismatch" in err_msg or "shape" in err_msg:
                logging.warning("Skipping model due to incompatible tensor shapes.")
                logging.warning("This typically indicates a flanking-size mismatch between the model and CLI arguments.")
                logging.warning(mismatch_hint)
                continue
            logging.error(f"Error processing model for device: {err_msg}. Skipping...")
            continue

        model = model.to(device)                # puts model on device
        model.eval()                            # puts model in evaluation mode
        loaded_models.append(model)             # appends model to list of loaded models  
            
    if not loaded_models:
        logging.error("No models were successfully loaded to the device.")
        exit()
        
    return loaded_models

def load_keras_models(model_path):
    """
    Loads Keras models from given path.
    
    Params:
    - model_path (str): Path to the model file or directory of models.
    
    Returns:
    - models (list): List of loaded Keras models.
    """
    from tensorflow import keras
    
    if os.path.isdir(model_path): # directory supplied
        model_files = glob.glob(os.path.join(model_path, '*.h5')) # get all Keras models from a directory
        if not model_files:
            logging.error(f"No Keras model files found in directory: {model_path}")
            exit()
            
        models = []
        for model_file in model_files:
            try:
                model = keras.models.load_model(model_file)
                models.append(model)
            except Exception as e:
                logging.error(f"Error loading Keras model from file {model_file}: {e}. Skipping...")
                
        if not models:
            logging.error(f"No valid PyTorch models found in directory: {model_path}")
            exit()
            
        return models
    
    elif os.path.isfile(model_path): # file supplied
        try:
            return [keras.models.load_model(model_path)]
        except Exception as e:
            logging.error(f"Error loading Keras model from file {model_path}: {e}")
            exit()
        
    else: # invalid path
        logging.error(f"Invalid path: {model_path}")
        exit()

##############################################
## FORMATTING INPUT DATA FOR PREDICTION
##############################################

def one_hot_encode(seq):
    """
    One-hot encode a DNA sequence.
    
    Args:
        seq (str): DNA sequence to be encoded.
    
    Returns:
        np.ndarray: One-hot encoded representation of the sequence.
    """

    # Define a mapping matrix for nucleotide to one-hot encoding
    map = np.asarray([[0, 0, 0, 0],  # N or any invalid character
                      [1, 0, 0, 0],  # A
                      [0, 1, 0, 0],  # C
                      [0, 0, 1, 0],  # G
                      [0, 0, 0, 1]]) # T

    # Replace nucleotides with corresponding indices
    seq = seq.upper().replace('A', '\x01').replace('C', '\x02')
    seq = seq.replace('G', '\x03').replace('T', '\x04').replace('N', '\x00')

    # Convert the sequence to one-hot encoded numpy array.
    # np.fromstring(binary) was removed in NumPy 2.0; frombuffer works across NumPy versions.
    return map[np.frombuffer(seq.encode('latin-1'), dtype=np.int8) % 5]


####################################################################################################################################
#######                                                                                                                      #######
#######                                             ANNOTATOR CLASS                                                          #######
#######                                                                                                                      #######
####################################################################################################################################

class Annotator:
    """
    Annotator class to handle gene annotations and reference sequences.
    It initializes with the reference genome, annotation data, and optional model configuration.
    """
    
    def __init__(self, ref_fasta, annotations, model_path='SpliceAI', model_type='keras', CL=80):
        """
        Initializes the Annotator with reference genome, annotations, and model settings.
        
        Args:
            ref_fasta (str): Path to the reference genome FASTA file.
            annotations (str): Path or name of the annotation file (e.g., 'grch37', 'grch38').
            model_path (str, optional): Path to the model file or type of model ('SpliceAI'). Defaults to SpliceAI.
            model_type (str, optional): Type of model ('keras' or 'pytorch'). Defaults to 'keras'.
            CL (int, optional): Context length parameter for model conversion. Defaults to 80.
        """

        # Load annotation file based on provided annotations type
        if annotations == 'grch37':
            annotations = './data/vcf/grch37.txt'
        elif annotations == 'grch38':
            annotations = './data/vcf/grch38.txt'

        # Load and parse the annotation file
        try:
            df = pd.read_csv(annotations, sep='\t', dtype={'CHROM': object})
            # Extract relevant columns into numpy arrays for efficient access
            self.genes = df['#NAME'].to_numpy()
            self.chroms = df['CHROM'].to_numpy()
            self.strands = df['STRAND'].to_numpy()
            self.tx_starts = df['TX_START'].to_numpy() + 1  # Transcription start sites (1-based indexing)
            self.tx_ends = df['TX_END'].to_numpy()  # Transcription end sites
            
            # Extract and process exon start and end sites, convert into numpy array format
            self.exon_starts = [np.asarray([int(i) for i in c.split(',') if i]) + 1
                                for c in df['EXON_START'].to_numpy()]
            self.exon_ends = [np.asarray([int(i) for i in c.split(',') if i])
                              for c in df['EXON_END'].to_numpy()]
        except IOError as e:
            logging.error('{}'.format(e)) 
            exit()  # Exit if the file cannot be read
        except (KeyError, pd.errors.ParserError) as e:
            logging.error('Gene annotation file {} not formatted properly: {}'.format(annotations, e))
            exit()  # Exit if the file format is incorrect

        # Load the reference genome fasta file
        try:
            self.ref_fasta = Fasta(ref_fasta, sequence_always_upper=True, rebuild=False)
        except IOError as e:
            logging.error('{}'.format(e))  # Log file read error
            exit()  # Exit if the file cannot be read

        # Load models based on the specified model type or file
        if model_path == 'SpliceAI':
            from tensorflow import keras
            paths = ('./models/spliceai/spliceai{}.h5'.format(x) for x in range(1, 6))  # Generate paths for SpliceAI models
            self.models = [keras.models.load_model(x) for x in paths]
            self.keras = True
        elif model_type == 'keras': # load models using keras
            self.models = load_keras_models(model_path)
            self.keras = True
        elif model_type == 'pytorch': # load models using pytorch 
            self.models = load_pytorch_models(model_path, CL)
            self.keras = False
        else:
            logging.error('Model type {} not supported'.format(model_type))
            exit()
        
        print(f'\t[INFO] {len(self.models)} model(s) loaded successfully')

    def get_name_and_strand(self, chrom, pos):
        """
        Retrieve gene names and strands overlapping a given chromosome position.
        
        Args:
            chrom (str): Chromosome identifier.
            pos (int): Position on the chromosome.
        
        Returns:
            tuple: Lists of gene names, strands, and their indices overlapping the given position.
        """

        # Normalize chromosome identifier to match the annotation format
        chrom = normalise_chrom(chrom, list(self.chroms)[0])
        # Find indices of annotations overlapping the given chromosome and position
        idxs = np.intersect1d(np.nonzero(self.chroms == chrom)[0],
                              np.intersect1d(np.nonzero(self.tx_starts <= pos)[0],
                                             np.nonzero(pos <= self.tx_ends)[0]))

        if len(idxs) >= 1:
            return self.genes[idxs], self.strands[idxs], idxs  # Return matching gene names and strands
        else:
            return [], [], []  # Return empty lists if no matches are found

    def get_pos_data(self, idx, pos):
        """
        Calculate distances from a given position to the transcription start site, 
        transcription end site, and nearest exon boundary for a specific gene.
        
        Args:
            idx (int): Index of the gene in the annotations.
            pos (int): Position on the chromosome.
        
        Returns:
            tuple: Distances to transcription start, transcription end, and nearest exon boundary.
        """

        # Calculate distances to transcription start and end sites
        dist_tx_start = self.tx_starts[idx] - pos
        dist_tx_end = self.tx_ends[idx] - pos
        # Calculate the closest distance to an exon boundary
        dist_exon_bdry = min(np.union1d(self.exon_starts[idx], self.exon_ends[idx]) - pos, key=abs)
        dist_ann = (dist_tx_start, dist_tx_end, dist_exon_bdry)  # Package distances into a tuple

        return dist_ann
    
##############################################
## CALCULATING DELTA SCORES
##############################################

def normalise_chrom(source, target):
    """
    Normalize chromosome identifiers to ensure consistency in format (with or without 'chr' prefix).
    
    Args:
        source (str): Source chromosome identifier.
        target (str): Target chromosome identifier for comparison.
    
    Returns:
        str: Normalized chromosome identifier.
    """

    def has_prefix(x):
        return x.startswith('chr')  # Check if a chromosome name has 'chr' prefix

    if has_prefix(source) and not has_prefix(target):
        return source.strip('chr')  # Remove 'chr' prefix if target doesn't have it
    elif not has_prefix(source) and has_prefix(target):
        return 'chr' + source  # Add 'chr' prefix if target has it

    return source  # Return source as is if both or neither have 'chr' prefix

def get_delta_scores(record, ann, dist_var, mask, flanking_size=10000, precision=2):
    """
    Calculate delta scores for variant impacts on splice sites.
    
    Args:
        record (pysam Record): Record containing variant information (e.g., chrom, pos, ref, alts).
        ann (Annotator): Annotator instance with annotation and reference genome data.
        dist_var (int): Max distance between variant and gained/lost splice site, defaults to 50.
        mask (bool): Mask scores representing annotated acceptor/donor gain and unannotated acceptor/donor loss, defaults to 0.
        flanking_size (int, optional): Size of the flanking region around the variant, defaults to 10000.
    
    Returns:
        list: Delta scores indicating the impact of variants on splicing.
    """

    # Define coverage and window size around the variant
    cov = 2 * dist_var + 1
    wid = flanking_size + cov
    delta_scores = []
    device = setup_device()

    # Validate the record fields
    try:
        record.chrom, record.pos, record.ref, len(record.alts)
    except TypeError:
        logging.warning('Skipping record (bad input): {}'.format(record))
        return delta_scores

    # Get gene names and strands overlapping the variant position
    (genes, strands, idxs) = ann.get_name_and_strand(record.chrom, record.pos)
    if len(idxs) == 0:
        return delta_scores  # Return empty list if no overlapping genes are found

    # Normalize chromosome and retrieve reference sequence around the variant
    chrom = normalise_chrom(record.chrom, list(ann.ref_fasta.keys())[0])
    try:
        seq = ann.ref_fasta[chrom][record.pos - wid // 2 - 1 : record.pos + wid // 2].seq
    except (IndexError, ValueError):
        logging.warning('Skipping record (fasta issue): {}'.format(record))
        return delta_scores

    # Check if the reference sequence matches the expected reference allele
    if seq[wid // 2 : wid // 2 + len(record.ref)].upper() != record.ref:
        logging.warning('Skipping record (ref issue): {}'.format(record))
        return delta_scores

    # Check if the sequence length matches the expected window size
    if len(seq) != wid:
        logging.warning('Skipping record (near chromosome end): {}'.format(record))
        return delta_scores

    # Skip records with a reference allele longer than the distance variable
    if len(record.ref) > 2 * dist_var:
        logging.warning('Skipping record (ref too long): {}'.format(record))
        return delta_scores

    # Iterate over each alternative allele and each gene index to calculate delta score
    for j in range(len(record.alts)):
        for i in range(len(idxs)):

            # Skip specific alternative allele types
            if '.' in record.alts[j] or '-' in record.alts[j] or '*' in record.alts[j]:
                continue
            if '<' in record.alts[j] or '>' in record.alts[j]:
                continue

            # Handle multi-nucleotide variants
            if len(record.ref) > 1 and len(record.alts[j]) > 1:
                delta_scores.append("{}|{}|.|.|.|.|.|.|.|.".format(record.alts[j], genes[i]))
                continue

            # Calculate position-related distances
            dist_ann = ann.get_pos_data(idxs[i], record.pos)
            pad_size = [max(wid // 2 + dist_ann[0], 0), max(wid // 2 - dist_ann[1], 0)]
            ref_len = len(record.ref)
            alt_len = len(record.alts[j])
            del_len = max(ref_len - alt_len, 0)

            # Construct reference and alternative sequences with padding
            x_ref = 'N' * pad_size[0] + seq[pad_size[0]: wid - pad_size[1]] + 'N' * pad_size[1]
            x_alt = x_ref[: wid // 2] + str(record.alts[j]) + x_ref[wid // 2 + ref_len:]

            # One-hot encode the sequences
            x_ref = one_hot_encode(x_ref)[None, :]
            x_alt = one_hot_encode(x_alt)[None, :]

            '''separate handling of PyTorch and Keras models'''
            if ann.keras: # keras model handling
                # Reverse the sequences if on the negative strand
                if strands[i] == '-':
                    x_ref = x_ref[:, ::-1, ::-1]
                    x_alt = x_alt[:, ::-1, ::-1]

                # Predict scores using the models
                y_ref = np.mean([ann.models[m].predict(x_ref) for m in range(len(ann.models))], axis=0)
                y_alt = np.mean([ann.models[m].predict(x_alt) for m in range(len(ann.models))], axis=0)
                
                # Reverse the predicted scores if on the negative strand
                if strands[i] == '-':
                    y_ref = y_ref[:, ::-1]
                    y_alt = y_alt[:, ::-1]
                    
            else: # pytorch model handling
                
                # Reshape tensor to match the model input shape
                x_ref = x_ref.transpose(0, 2, 1)
                x_alt = x_alt.transpose(0, 2, 1)
                
                # Convert to PyTorch tensors
                x_ref = torch.tensor(x_ref, dtype=torch.float32)
                x_alt = torch.tensor(x_alt, dtype=torch.float32)

                # Reverse the sequences if on the negative strand
                if strands[i] == '-':
                    x_ref = torch.flip(x_ref, dims=[1, 2])
                    x_alt = torch.flip(x_alt, dims=[1, 2])

                # Put tensors on device
                x_ref = x_ref.to(device)
                x_alt = x_alt.to(device)
                
                # Predict scores using the models
                with torch.no_grad():
                    y_ref = torch.mean(torch.stack([ann.models[m](x_ref).detach().cpu() for m in range(len(ann.models))]), axis=0)
                    y_alt = torch.mean(torch.stack([ann.models[m](x_alt).detach().cpu() for m in range(len(ann.models))]), axis=0)
                
                # Remove flanking sequence and permute shape
                y_ref = y_ref.permute(0, 2, 1)
                y_alt = y_alt.permute(0, 2, 1)

                # Reverse the predicted scores if on the negative strand and convert back to numpy arrays
                if strands[i] == '-':
                    y_ref = torch.flip(y_ref, dims=[1])
                    y_alt = torch.flip(y_alt, dims=[1])
                
                # Convert to numpy arrays
                y_ref = y_ref.numpy()
                y_alt = y_alt.numpy()
            '''end'''

            # Manually crop the output if it exceeds the expected coverage window
            # This handles cases where the model output is not automatically cropped to the target window
            if y_ref.shape[1] > cov:
                start_idx = wid // 2 - cov // 2
                y_ref = y_ref[:, start_idx : start_idx + cov, :]
                y_alt = y_alt[:, start_idx : start_idx + cov + alt_len - ref_len, :]


            # Adjust the alternative sequence scores based on reference and alternative lengths
            if ref_len > 1 and alt_len == 1:
                y_alt = np.concatenate([
                    y_alt[:, : cov // 2 + alt_len],
                    np.zeros((1, del_len, 3)),
                    y_alt[:, cov // 2 + alt_len:]
                ], axis=1)
            elif ref_len == 1 and alt_len > 1:
                y_alt = np.concatenate([
                    y_alt[:, : cov // 2],
                    np.max(y_alt[:, cov // 2 : cov // 2 + alt_len], axis=1)[:, None, :],
                    y_alt[:, cov // 2 + alt_len:]
                ], axis=1)

            # Concatenate the reference and alternative scores
            y = np.concatenate([y_ref, y_alt])

            # Find the indices of maximum delta scores for splicing acceptor and donor sites
            idx_pa = (y[1, :, 1] - y[0, :, 1]).argmax()
            idx_na = (y[0, :, 1] - y[1, :, 1]).argmax()
            idx_pd = (y[1, :, 2] - y[0, :, 2]).argmax()
            idx_nd = (y[0, :, 2] - y[1, :, 2]).argmax()

            # print(f"idx_pa: {idx_pa}, idx_na: {idx_na}, idx_pd: {idx_pd}, idx_nd: {idx_nd}")
            # print("cov:", cov)

            # Apply masks to delta scores based on calculated indices and provided mask
            mask_pa = np.logical_and((idx_pa - cov // 2 == dist_ann[2]), mask)
            mask_na = np.logical_and((idx_na - cov // 2 != dist_ann[2]), mask)
            mask_pd = np.logical_and((idx_pd - cov // 2 == dist_ann[2]), mask)
            mask_nd = np.logical_and((idx_nd - cov // 2 != dist_ann[2]), mask)

            # Create a format string with the desired precision
            format_str = "{{}}|{{}}|{{:.{}f}}|{{:.{}f}}|{{:.{}f}}|{{:.{}f}}|{{}}|{{}}|{{}}|{{}}".format(
                precision, precision, precision, precision)
            
            # Write delta scores for given alternative allele, gene, and calculated indices
            delta_scores.append(format_str.format(
                record.alts[j],
                genes[i],
                (y[1, idx_pa, 1] - y[0, idx_pa, 1]) * (1 - mask_pa),
                (y[0, idx_na, 1] - y[1, idx_na, 1]) * (1 - mask_na),
                (y[1, idx_pd, 2] - y[0, idx_pd, 2]) * (1 - mask_pd),
                (y[0, idx_nd, 2] - y[1, idx_nd, 2]) * (1 - mask_nd),
                idx_pa - cov // 2,
                idx_na - cov // 2,
                idx_pd - cov // 2,
                idx_nd - cov // 2
            ))

    return delta_scores 


import argparse
import gffutils
import os

def parse_arguments():
    parser = argparse.ArgumentParser(description="Convert GFF file to custom tabular format.")
    parser.add_argument(
        "--gff",
        required=True,
        help="Path to the input GFF file (e.g., data/GRCh38_chr21_chr22.gff)."
    )
    parser.add_argument(
        "--output",
        required=True,
        help="Path to the output custom annotation file (e.g., data/grch38_chr21_chr22.txt)."
    )
    parser.add_argument(
        "--db",
        default=None,
        help="Path to the temporary GFF database. If not provided, a temporary database will be created."
    )
    return parser.parse_args()

def create_gff_database(gff_file, db_file=None):
    """
    Create a gffutils database from the GFF file.
    If db_file is None, a temporary in-memory database is created.
    """
    if db_file and os.path.exists(db_file):
        print(f"Connecting to existing GFF database: {db_file}")
        db = gffutils.FeatureDB(db_file)
    else:
        if db_file:
            print(f"Creating new GFF database: {db_file}")
        else:
            print("Creating in-memory GFF database.")
        db = gffutils.create_db(
            gff_file,
            dbfn=db_file if db_file else ":memory:",
            force=True,
            keep_order=True,
            merge_strategy='create_unique',
            sort_attribute_values=True,
            verbose=True
        )
    return db

def extract_gene_info(db):
    """
    Extract gene information from the GFF database.
    Returns a list of dictionaries containing gene details.
    """
    gene_data = []
    for gene in db.features_of_type('gene'):
        # extract gene name
        gene_name = None
        for key in ['gene_name', 'Name', 'gene', 'ID']:
            if key in gene.attributes:
                gene_name = gene.attributes[key][0]
                break
        if not gene_name:
            gene_name = gene.id  # fallback to gene ID if name not available

        chrom = gene.seqid
        strand = gene.strand
        tx_start = gene.start
        tx_end = gene.end

        # Extract exons
        exons = list(db.children(gene, featuretype='exon', order_by='start'))
        if not exons:
            continue  # Skip genes without exons

        exon_starts = []
        exon_ends = []
        for exon in exons:
            exon_starts.append(str(exon.start))
            exon_ends.append(str(exon.end))
        
        # Concatenate exon positions with commas
        exon_start_str = ",".join(exon_starts) + ","
        exon_end_str = ",".join(exon_ends) + ","

        gene_entry = {
            "NAME": gene_name,
            "CHROM": chrom,
            "STRAND": strand,
            "TX_START": tx_start,
            "TX_END": tx_end,
            "EXON_START": exon_start_str,
            "EXON_END": exon_end_str
        }
        gene_data.append(gene_entry)
    return gene_data

def write_output(gene_data, output_file):
    """
    Write the extracted gene information to the output file in tab-separated format.
    """
    header = ["#NAME", "CHROM", "STRAND", "TX_START", "TX_END", "EXON_START", "EXON_END"]
    with open(output_file, "w") as out_f:
        out_f.write("\t".join(header) + "\n")
        for gene in gene_data:
            row = [
                gene["NAME"],
                gene["CHROM"],
                gene["STRAND"],
                str(gene["TX_START"]),
                str(gene["TX_END"]),
                gene["EXON_START"],
                gene["EXON_END"]
            ]
            out_f.write("\t".join(row) + "\n")
    print(f"Output written to {output_file}")

def main():
    args = parse_arguments()

    # Create or connect to GFF database
    db = create_gff_database(args.gff, args.db)

    # Extract gene information
    gene_data = extract_gene_info(db)

    # Write to output file
    write_output(gene_data, args.output)

if __name__ == "__main__":
    main()


"""
Filename: create_datafile.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Create datafile.h5 for training and testing splice site prediction models.
"""

import os
from Bio.Seq import Seq  # Import the Seq class
from Bio import SeqIO, SeqRecord
import time
import openspliceai.create_data.utils as utils
import openspliceai.create_data.paralogs as paralogs

donor_motif_counts = {}  # Initialize counts
acceptor_motif_counts = {}  # Initialize counts

def get_sequences_and_labels(db, output_dir, seq_dict, chrom_dict, train_or_test, parse_type="canonical", biotype="protein-coding", canonical_only=True, write_fasta=False):
    """
    Extract sequences for each protein-coding gene, reverse complement sequences for genes on the reverse strand,
    and label donor and acceptor sites correctly based on strand orientation.
    """
    fw_stats = open(os.path.join(output_dir, f"{train_or_test}_stats.txt"), "w")
    if write_fasta:
        fasta_handle = open(os.path.join(output_dir, f"{train_or_test}.fa"), "w")  # Open a file to write FASTA format sequences
    NAME = []      # Gene Name
    CHROM = []     # Chromosome
    STRAND = []    # Strand in which the gene lies (+ or -)
    TX_START = []  # Position where transcription starts
    TX_END = []    # Position where transcription ends
    SEQ = []       # Nucleotide sequence
    LABEL = []     # Label for each nucleotide in the sequence
    for gene in db.features_of_type('gene'):
        if "exception" in gene.attributes.keys() and gene.attributes["exception"][0] == "trans-splicing":
            continue
        if gene.seqid not in chrom_dict:
            continue
        if biotype == "protein-coding":
            if gene.attributes["gene_biotype"][0] != "protein_coding":
                continue
        elif biotype == "non-coding":
            if gene.attributes["gene_biotype"][0] != "lncRNA" and gene.attributes["gene_biotype"][0] != "ncRNA":
                continue
        elif biotype == "all":
            if gene.attributes["gene_biotype"][0] != "protein_coding" and gene.attributes["gene_biotype"][0] != "lncRNA" and gene.attributes["gene_biotype"][0] != "ncRNA":
                continue
        else:
            continue
        chrom_dict[gene.seqid] += 1
        gene_id = gene.id

        # Process the gene sequence
        gene_seq = seq_dict[gene.seqid].seq[gene.start-1:gene.end].upper()  # Extract gene sequence
        if gene.strand == '-':
            gene_seq = gene_seq.reverse_complement() # reverse complement the sequence
        gene_seq = str(gene_seq.upper())
        print(f"\tProcessing gene {gene_id} on chromosome {gene.seqid}..., len(gene_seq): {len(gene_seq)}")
        labels = [0] * len(gene_seq)  # Initialize all labels to 0
        if biotype =="protein-coding":
            transcripts = list(db.children(gene, featuretype='mRNA', order_by='start'))
        elif biotype =="non-coding":
            transcripts = list(db.children(gene, level=1, order_by='start'))
        if len(transcripts) == 0:
            continue
        elif len(transcripts) > 1:
            print(f"\tGene {gene_id} has multiple transcripts: {len(transcripts)}")

        # Selecting which mode to process the data
        transcripts_ls = []
        if parse_type == 'canonical':
            max_trans = transcripts[0]
            max_len = max_trans.end - max_trans.start + 1
            for transcript in transcripts:
                if transcript.end - transcript.start + 1 > max_len:
                    max_trans = transcript
                    max_len = transcript.end - transcript.start + 1
            transcripts_ls = [max_trans]
        elif parse_type == 'all_isoforms':
            transcripts_ls = transcripts
        
        # Process transcripts
        for transcript in transcripts_ls:
            exons = list(db.children(transcript, featuretype='exon', order_by='start'))
            if len(exons) > 1:
                for i in range(len(exons) - 1):
                    # Donor site is one base after the end of the current exon
                    first_site = exons[i].end - gene.start  # Adjusted for python indexing
                    # Acceptor site is at the start of the next exon
                    second_site = exons[i + 1].start - gene.start          
                    # Adjusted for python indexing
                    if gene.strand == '+':
                        d_idx = first_site
                        a_idx = second_site
                    elif gene.strand == '-':
                        d_idx = len(labels) - second_site-1
                        a_idx = len(labels) - first_site-1
                    # Check if the donor / acceptor sites are valid
                    d_motif = str(gene_seq[d_idx+1:d_idx+3])
                    a_motif = str(gene_seq[a_idx-2:a_idx])
                    if not canonical_only:
                        labels[d_idx] = 2   # Mark donor site
                        labels[a_idx] = 1  # Mark acceptor site
                    elif (d_motif == "GT" and a_motif == "AG") or (d_motif == "GC" and a_motif == "AG") or (d_motif == "AT" and a_motif == "AC"):
                        labels[d_idx] = 2   # Mark donor site
                        labels[a_idx] = 1  # Mark acceptor site
            labels_str = ''.join(str(num) for num in labels)
            NAME.append(gene_id)
            CHROM.append(gene.seqid)
            STRAND.append(gene.strand)
            TX_START.append(str(gene.start))
            TX_END.append(str(gene.end))
            SEQ.append(gene_seq)
            LABEL.append(labels_str)

            # Write to the FASTA file if the path is provided
            if write_fasta:
                record = SeqRecord.SeqRecord(
                    Seq(gene_seq),
                    id=gene_id,
                    description=f"{gene.seqid}:{gene.start}-{gene.end}({gene.strand})"
                )
                SeqIO.write(record, fasta_handle, "fasta")

            fw_stats.write(f"{gene.seqid}\t{gene.start}\t{gene.end}\t{gene.id}\t{1}\t{gene.strand}\n")
            utils.check_and_count_motifs(gene_seq, labels, donor_motif_counts, acceptor_motif_counts)
    print(f"Total SEQ: {len(SEQ)}")
    fw_stats.close()
    if write_fasta:
        fasta_handle.close()  # Close the FASTA file handle
    return [NAME, CHROM, STRAND, TX_START, TX_END, SEQ, LABEL]


def create_datafile(args):
    print("Running OpenSpliceAI with 'create-data' mode")
    print("--- Step 1: Creating datafile.h5 ... ---")
    start_time = time.process_time()
    
    # Use gffutils to parse annotation file
    os.makedirs(args.output_dir, exist_ok=True)
    db = utils.create_or_load_db(args.annotation_gff, db_file=f'{args.annotation_gff}_db')
    seq_dict = SeqIO.to_dict(SeqIO.parse(args.genome_fasta, "fasta"))
        
    # Find all distinct chromosomes and split them
    TRAIN_CHROM_GROUP, TEST_CHROM_GROUP = utils.split_chromosomes(seq_dict, method=args.split_method, split_ratio=args.split_ratio)
    print("* TRAIN_CHROM_GROUP: ", TRAIN_CHROM_GROUP)
    print("* TEST_CHROM_GROUP: ", TEST_CHROM_GROUP)

    # Collect sequences and labels for testing and/or training groups
    if args.chr_split == 'test':
        print("> Creating test datafile...")
        test_data = get_sequences_and_labels(db, args.output_dir, seq_dict, TEST_CHROM_GROUP, 'test', parse_type=args.parse_type, biotype=args.biotype, canonical_only=args.canonical_only, write_fasta=args.write_fasta)
        paralogs.write_h5_file(args.output_dir, "test", test_data)
    elif args.chr_split == 'train-test':
        print("> Creating train datafile...")
        train_data = get_sequences_and_labels(db, args.output_dir, seq_dict, TRAIN_CHROM_GROUP, 'train', parse_type=args.parse_type, biotype=args.biotype, canonical_only=args.canonical_only, write_fasta=args.write_fasta)
        print("> Creating test datafile...")
        test_data = get_sequences_and_labels(db, args.output_dir, seq_dict, TEST_CHROM_GROUP, 'test', parse_type=args.parse_type, biotype=args.biotype, canonical_only=args.canonical_only, write_fasta=args.write_fasta)
        if args.remove_paralogs:
            # Remove homologous sequences
            print("> Removing homologous sequences in test...")
            train_data, test_data = paralogs.remove_paralogous_sequences(train_data, test_data, args.min_identity, args.min_coverage, args.output_dir, "test")

        # Split the training data into training and validation sets (90:10 split)
        print("> Splitting training data into training and validation sets...")
        train_data, val_data = utils.split_train_val(train_data, args.val_split_ratio)
        if args.remove_paralogs:
            # Remove homologous sequences
            print("> Removing homologous sequences in validation...")
            train_data, val_data = paralogs.remove_paralogous_sequences(train_data, val_data, args.min_identity, args.min_coverage, args.output_dir, "validation")

        # Write the filtered data to h5 files
        paralogs.write_h5_file(args.output_dir, "train", train_data)
        paralogs.write_h5_file(args.output_dir, "validation", val_data)
        paralogs.write_h5_file(args.output_dir, "test", test_data)
    utils.print_motif_counts(donor_motif_counts, acceptor_motif_counts)    
    print("--- %s seconds ---" % (time.process_time() - start_time))



"""
Filename: train.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Train the OpenSpliceAI model.
"""

import sys
import numpy as np
import torch
import torch.optim as optim
from torch.utils.data import TensorDataset, DataLoader
from tqdm import tqdm
from openspliceai.train_base.openspliceai import *
from openspliceai.train_base.utils import *
from openspliceai.constants import *
import openspliceai.create_data.paralogs as paralogs

def initialize_model_and_optim(device, flanking_size, epochs, scheduler):
    # Hyper-parameters:
    # L: Number of convolution kernels
    # W: Convolution window size in each residual unit
    # AR: Atrous rate in each residual unit
    L = 32
    N_GPUS = 2
    W = np.asarray([11, 11, 11, 11])
    AR = np.asarray([1, 1, 1, 1])
    BATCH_SIZE = 18*N_GPUS
    if int(flanking_size) == 80:
        W = np.asarray([11, 11, 11, 11])
        AR = np.asarray([1, 1, 1, 1])
        BATCH_SIZE = 18*N_GPUS
    elif int(flanking_size) == 400:
        W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11])
        AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4])
        BATCH_SIZE = 18*N_GPUS
    elif int(flanking_size) == 2000:
        W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11,
                        21, 21, 21, 21])
        AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4,
                        10, 10, 10, 10])
        BATCH_SIZE = 12*N_GPUS
    elif int(flanking_size) == 10000:
        W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11,
                        21, 21, 21, 21, 41, 41, 41, 41])
        AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4,
                        10, 10, 10, 10, 25, 25, 25, 25])
        BATCH_SIZE = 6*N_GPUS
    CL = 2 * np.sum(AR*(W-1))
    print("\033[1mContext nucleotides: %d\033[0m" % (CL))
    print("\033[1mSequence length (output): %d\033[0m" % (SL))
    model = SpliceAI(L, W, AR).to(device)
    print(model, file=sys.stderr)
    # optimizer = optim.Adam(model.parameters(), lr=1e-3)
    optimizer = torch.optim.AdamW(model.parameters(), lr=1e-3)
    # scheduler = get_cosine_schedule_with_warmup(optimizer, 1000, train_size * EPOCH_NUM)
    if scheduler == "MultiStepLR":
        scheduler_obj = optim.lr_scheduler.MultiStepLR(optimizer, milestones=[epochs-5, epochs-4, epochs-3, epochs-2, epochs-1], gamma=0.5)
    elif scheduler == "CosineAnnealingWarmRestarts":
        scheduler_obj = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(
            optimizer, T_0=5, T_mult=1, eta_min=1e-5, last_epoch=-1)    
    params = {'L': L, 'W': W, 'AR': AR, 'CL': CL, 'SL': SL, 'BATCH_SIZE': BATCH_SIZE, 'N_GPUS': N_GPUS}
    return model, optimizer, scheduler_obj, params


def train(args):
    print("Running OpenSpliceAI with 'train' mode")
    device = setup_environment(args)
    model_output_base, log_output_train_base, log_output_val_base, log_output_test_base = initialize_paths(args)
    train_h5f, valid_h5f, test_h5f, batch_num = load_datasets(args)
    train_idxs, val_idxs, test_idxs = generate_indices(train_h5f, valid_h5f, test_h5f)

    model, optimizer, scheduler, params = initialize_model_and_optim(device, args.flanking_size, args.epochs, args.scheduler)
    params["RANDOM_SEED"] = args.random_seed
    train_metric_files = create_metric_files(log_output_train_base)
    valid_metric_files = create_metric_files(log_output_val_base)
    test_metric_files = create_metric_files(log_output_test_base)
    train_model(model, optimizer, scheduler, train_h5f, valid_h5f, test_h5f, 
                train_idxs, val_idxs, test_idxs, model_output_base, args, device, params, train_metric_files, valid_metric_files, test_metric_files)
    train_h5f.close()
    valid_h5f.close()
    test_h5f.close()


"""
Filename: visualization.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Functions for visualizing calibration results.
"""

import matplotlib.pyplot as plt
import numpy as np
from openspliceai.calibrate.calibrate_utils import *
import torch

def plot_score_distribution(probs, probs_scaled, labels, output_dir, index):
    index_names = {
        0: ('non-splice site', 'neither'),
        1: ('acceptor site', 'acceptor'),
        2: ('donor site', 'donor')
    }
    
    positive_indices = (labels == index)
    probs_positive = probs[:, index][positive_indices]
    probs_scaled_positive = probs_scaled[:, index][positive_indices]

    plt.figure(figsize=(5.5, 4))
    plt.hist(probs_positive, bins=30, alpha=0.5, label='Original')
    plt.hist(probs_scaled_positive, bins=30, alpha=0.5, label='Calibrated')
    plt.xlabel('Score')
    plt.ylabel('Density')
    plt.legend()
    plt.grid(True)
    plt.title(f'Score Distribution for {index_names[index][0]}')
    plt.tight_layout()
    plt.savefig(f"{output_dir}/prob_dist_{index_names[index][1]}.png")
    plt.close()


def score_frequency_distribution(probs, probs_scaled, labels, outdir, index=1):
    # Flatten the tensors for simplicity
    probabilities_flat = probs[:, index]
    probabilities_scaled_flat = probs_scaled[:, index]
    print("probabilities_flat.shape: ", probabilities_flat.shape)
    print("probabilities_scaled_flat.shape: ", probabilities_scaled_flat.shape)
    print("labels.shape: ", labels.shape)
    # Filter out cases where the label is zero (i.e., keep only positive cases for Class 1)
    positive_indices = (labels == index)  # True for positive cases
    print("positive_indices: ", positive_indices)
    print("len(positive_indices): ", len(positive_indices))
    probabilities_positive = probabilities_flat[positive_indices]
    probabilities_scaled_positive = probabilities_scaled_flat[positive_indices]
    labels_positive = labels[positive_indices]
    print("len(probabilities_positive): ", len(probabilities_positive))
    print("probabilities_positive: ", probabilities_positive)
    print("len(probabilities_scaled_positive): ", len(probabilities_scaled_positive))
    print("probabilities_scaled_positive: ", probabilities_scaled_positive)
    print("len(labels_positive): ", len(labels_positive))
    print("labels_positive: ", labels_positive)
    # Since we're now only looking at positive cases, the labels will be all ones,
    # and we're primarily interested in the distribution of predicted probabilities.
    # Plotting
    plt.figure(figsize=(5.5, 4))
    # Probability distribution for Class 1 from logits (only positive cases)
    plt.hist(probabilities_positive, bins=30, alpha=0.5, label='Predicted Probabilities')
    plt.hist(probabilities_scaled_positive, bins=30, alpha=0.5, label='Predicted Calibrated Probabilities')
    # For positive cases, since all labels are 1, we can simply indicate the average or density with a line or annotation
    # plt.axvline(x=np.mean(labels_positive), color='r', linestyle='dashed', linewidth=2, label='Actual Labels')
    plt.xlabel('Score')
    plt.ylabel('Density')
    plt.legend()
    plt.grid(True)
    if index == 0:
        plt.title('Score Distribution for non-splice site')
        plt.tight_layout()
        plt.savefig(f"{outdir}/prob_dist_neither.png")
    if index == 1:
        plt.title('Score Distribution for acceptor site')
        plt.tight_layout()
        plt.savefig(f"{outdir}/prob_dist_acceptor.png")
    elif index == 2:
        plt.title('Score Distribution for donor site')
        plt.tight_layout()
        plt.savefig(f"{outdir}/prob_dist_donor.png")
    plt.clf()


def plot_calibration_curves(calibration_data, calibration_data_scaled, classes, output_dir):
    fig, axs = plt.subplots(1, 3, figsize=(20, 5.5))
    for i, (class_name, (orig_data, scaled_data)) in enumerate(zip(classes, zip(calibration_data, calibration_data_scaled))):
        prob_true, prob_pred, bin_counts = orig_data
        prob_true_s, prob_pred_s, _ = scaled_data
        
        ci_lower, ci_upper = compute_confidence_intervals(prob_true, bin_counts)
        ci_lower_s, ci_upper_s = compute_confidence_intervals(prob_true_s, bin_counts)
        
        axs[i].plot(prob_pred, prob_true, 'o-', color='blue', label='Original')
        axs[i].fill_between(prob_pred, ci_lower, ci_upper, color='blue', alpha=0.1)
        axs[i].plot(prob_pred_s, prob_true_s, 'o-', color='green', label='Calibrated')
        axs[i].fill_between(prob_pred_s, ci_lower_s, ci_upper_s, color='green', alpha=0.1)
        axs[i].plot([0, 1], [0, 1], 'k--')
        axs[i].set_title(class_name)
        axs[i].set_xlabel('Predicted Probability')
        axs[i].set_ylabel('Empirical Probability')
        axs[i].legend()
    plt.tight_layout()
    plt.savefig(f"{output_dir}/calibration_curve.png", dpi=300)
    plt.close()


def plot_brier_scores(brier_uncal, brier_cal, classes, output_dir):
    plt.figure(figsize=(8, 6))
    x = np.arange(len(classes))
    plt.bar(x - 0.2, brier_uncal, 0.4, label='Uncalibrated')
    plt.bar(x + 0.2, brier_cal, 0.4, label='Calibrated')
    plt.xticks(x, classes)
    plt.ylabel('Brier Score')
    plt.title('Brier Score Comparison')
    plt.legend()
    plt.tight_layout()
    plt.savefig(f"{output_dir}/brier_scores.png", dpi=300)
    plt.close()


def plot_calibration_map(scaled_model, device, output_dir):
    p1d = np.linspace(0, 1, 20)
    p0, p1 = np.meshgrid(p1d, p1d)
    p2 = 1 - p0 - p1
    mask = (p2 >= 0) & (p2 <= 1)
    p = np.vstack([p0[mask], p1[mask], p2[mask]]).T
    
    logits = reverse_softmax(p)
    with torch.no_grad():
        logits_scaled = scaled_model.temperature_scale(torch.tensor(logits).float().to(device))
    probs_scaled = torch.nn.functional.softmax(logits_scaled, dim=1).cpu().numpy()

    plt.figure(figsize=(4.8, 4.8))
    colors = ['red', 'green', 'blue']
    for i in range(p.shape[0]):
        color = colors[np.argmax(p[i])]
        plt.arrow(p[i][1], p[i][2], 
                 probs_scaled[i][1]-p[i][1], probs_scaled[i][2]-p[i][2],
                 color=color, head_width=0.01, length_includes_head=True)
    
    plt.plot([0, 1], [0, 0], 'k-')
    plt.plot([1, 0], [0, 1], 'k-')
    plt.plot([0, 0], [1, 0], 'k-')
    plt.xlabel("Acceptor Probability")
    plt.ylabel("Donor Probability")
    plt.title("Calibration Map")
    plt.legend(handles=[
        plt.Line2D([0], [0], color='red', label='Non-splice'),
        plt.Line2D([0], [0], color='green', label='Acceptor'),
        plt.Line2D([0], [0], color='blue', label='Donor')
    ])
    plt.tight_layout()
    plt.savefig(f"{output_dir}/calibration_map.png", dpi=300)
    plt.close()


"""
Filename: utils.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Functions to process sequences to/from .h5 datasets.
"""

import os
import numpy as np
from math import ceil
import gffutils
import random
from sklearn.metrics import average_precision_score
from openspliceai.constants import *

def check_and_count_motifs(seq, labels, donor_motif_counts, acceptor_motif_counts):
    """
    Check sequences for donor and acceptor motifs and count their occurrences.
    """     
    for i, label in enumerate(labels):
        if label == 2:  # Donor site
            d_motif = str(seq[i+1:i+3])
            donor_motif_counts[d_motif] = donor_motif_counts.get(d_motif, 0) + 1
        elif label == 1:  # Acceptor site
            a_motif = str(seq[i-2:i])
            acceptor_motif_counts[a_motif] = acceptor_motif_counts.get(a_motif, 0) + 1


def print_motif_counts(donor_motif_counts, acceptor_motif_counts):
    """
    Print the counts of donor and acceptor motifs.
    """
    print("*******************")
    print("Splice site motif counts:")
    print("\tDonor motifs:")
    for motif, count in donor_motif_counts.items():
        print(f"\t{motif}: {count}")
    print("\n\tAcceptor motifs:")
    for motif, count in acceptor_motif_counts.items():
        print(f"\t{motif}: {count}")
    print("\nTotal donor motifs: ", sum(donor_motif_counts.values()))
    print("Total acceptor motifs: ", sum(acceptor_motif_counts.values()))
    print("*******************")
    
    
###################################################
# create_datafile.py functions
###################################################
def get_chromosome_lengths(seq_dict):
    """
    Extract all unique chromosomes and their lengths
    """
    chrom_lengths = {chrom: len(record.seq) for chrom, record in seq_dict.items()}
    return chrom_lengths


def split_chromosomes(seq_dict, method='random', split_ratio=0.8):
    """
    Split chromosomes into training and testing groups.
    """
    chromosome_lengths = get_chromosome_lengths(seq_dict)
    print("Chromosome lengths: ", chromosome_lengths)
    if method == 'random':
        total_length = sum(chromosome_lengths.values())
        target_train_length = total_length * split_ratio
        target_test_length = total_length * (1-split_ratio)
        chromosomes = list(chromosome_lengths.keys())
        random.shuffle(chromosomes)
        train_chroms = {}
        test_chroms = {}
        current_train_length = 0
        current_test_length = 0 
        
        for chrom in chromosomes:
            if current_test_length < target_test_length:
                test_chroms[chrom] = chromosome_lengths[chrom]
                current_test_length += chromosome_lengths[chrom]
            else:
                train_chroms[chrom] = chromosome_lengths[chrom]
                current_train_length += chromosome_lengths[chrom]
    elif method == 'human':
        # following SpliceAI default splitting for human chromosomes
        train_chroms = {
        'chr2': 0, 'chr4': 0, 'chr6': 0, 'chr8': 0, 
        'chr10': 0, 'chr11': 0, 'chr12': 0, 'chr13': 0,
        'chr14': 0, 'chr15': 0, 'chr16': 0, 'chr17': 0, 
        'chr18': 0, 'chr19': 0, 'chr20': 0, 'chr21': 0, 
        'chr22': 0, 'chrX': 0, 'chrY': 0
        }
        test_chroms = {
            'chr1': 0, 'chr3': 0, 'chr5': 0, 'chr7': 0, 'chr9': 0
        }
    else: 
        raise ValueError("Invalid chromosome split method. Use 'random' or 'human'.")
    return train_chroms, test_chroms


def create_or_load_db(gff_file, db_file='gff.db'):
    """
    Create a gffutils database from a GFF file, or load it if it already exists.
    """
    if not os.path.exists(db_file):
        print("Creating new database...")
        db = gffutils.create_db(gff_file, dbfn=db_file, force=True, keep_order=True, merge_strategy='merge', sort_attribute_values=True)
    else:
        print("Loading existing database...")
        db = gffutils.FeatureDB(db_file)
    return db

###################################################
# create_dataset.py functions
###################################################
# One-hot encoding of the inputs: 
# 1: A;  2: C;  3: G;  4: T;  0: padding
IN_MAP = np.asarray([[0, 0, 0, 0],
                     [1, 0, 0, 0],
                     [0, 1, 0, 0],
                     [0, 0, 1, 0],
                     [0, 0, 0, 1]])

# One-hot encoding of the outputs: 
# 0: no splice;  1: acceptor;  2: donor;  -1: padding
OUT_MAP = np.asarray([[1, 0, 0],
                      [0, 1, 0],
                      [0, 0, 1],
                      [0, 0, 0]])


def ceil_div(x, y):
    """
    Calculate the ceiling of a division between two numbers.
    """
    return int(ceil(float(x)/y))


def one_hot_encode(Xd, Yd):
    """
    Perform one-hot encoding on both the input sequence data (Xd) 
    and the output label data (Yd) using predefined mappings 
    (IN_MAP for inputs and OUT_MAP for outputs).
    """
    return IN_MAP[Xd.astype('int8')], [OUT_MAP[Yd[t].astype('int8')] for t in range(1)]


def replace_non_acgt_to_n(input_string):
    """
    Use a generator expression to go through each character in the input string.
    If the character is in the set of allowed characters, keep it as is.
    Otherwise, replace it with 'N'.
    """
    # Define the set of allowed characters
    allowed_chars = {'A', 'C', 'G', 'T'}    
    return ''.join(char if char in allowed_chars else 'N' for char in input_string)


def reformat_data(X0, Y0):
    """
    Reformat sequence and label data into fixed-size blocks for processing.
    This function converts X0, Y0 of the create_datapoints function into
    blocks such that the data is broken down into data points where the
    input is a sequence of length SL+CL_max corresponding to SL nucleotides
    of interest and CL_max context nucleotides, the output is a sequence of
    length SL corresponding to the splicing information of the nucleotides
    of interest. The CL_max context nucleotides are such that they are
    CL_max/2 on either side of the SL nucleotides of interest.
    """
    # Calculate the number of data points needed
    num_points = ceil_div(len(Y0[0]), SL)
    # Initialize arrays to hold the reformatted data
    Xd = np.zeros((num_points, SL + CL_max))
    Yd = [-np.ones((num_points, SL)) for _ in range(1)]
    # Pad the sequence and labels to ensure divisibility
    X0 = np.pad(X0, (0, SL), 'constant', constant_values=0)
    Y0 = [np.pad(Y0[t], (0, SL), 'constant', constant_values=-1) for t in range(1)]
    # Fill the initialized arrays with data in blocks
    for i in range(num_points):
        Xd[i] = X0[SL * i : SL * (i + 1) + CL_max]
        Yd[0][i] = Y0[0][SL * i : SL * (i + 1)]

    return Xd, Yd  


def create_datapoints(seq, label):
    """
    This function first converts the sequence into an integer array, where
    A, C, G, T, N are mapped to 1, 2, 3, 4, 0 respectively. If the strand is
    negative, then reverse complementing is done. The labels 
    are directly used as they are, converted into an array of integers,
    where 0, 1, 2 correspond to no splicing, acceptor, donor 
    respectively. It then calls reformat_data and one_hot_encode
    and returns X, Y which can be used by Pytorch Model.
    """
    # No need to reverse complement the sequence, as sequence is already reverse complemented from previous step
    seq = 'N' * (CL_max // 2) + seq + 'N' * (CL_max // 2)
    seq = seq.upper().replace('A', '1').replace('C', '2')
    seq = seq.replace('G', '3').replace('T', '4').replace('N', '0')

    # Convert label string to array of integers
    label_array = np.array(list(map(int, list(label))))
    X0 = np.asarray(list(map(int, list(seq))))
    Y0 = label_array
    Y0 = [Y0]
    Xd, Yd = reformat_data(X0, Y0)
    X, Y = one_hot_encode(Xd, Yd)

    return X, Y


def split_train_val(data, val_split_ratio):
    """
    Split training data into training and validation sets.

    Args:
        data (list of lists): [NAME, CHROM, STRAND, TX_START, TX_END, SEQ, LABEL]
        val_split_ratio (float): proportion of data to use for validation (e.g., 0.1 for 10%)

    Returns:
        tuple: (train_data, val_data), each a list of lists with the same structure as input
    """
    # Transpose data to a list of records
    records = list(zip(*data))
    # Shuffle records in place for random split
    random.shuffle(records)
    total = len(records)
    val_size = int(total * val_split_ratio)

    # Split into validation and training records
    val_records = records[:val_size]
    train_records = records[val_size:]
    # Handle case where no records
    if not train_records:
        train_data = [[] for _ in data]
    else:
        train_data = [list(field) for field in zip(*train_records)]
    if not val_records:
        val_data = [[] for _ in data]
    else:
        val_data = [list(field) for field in zip(*val_records)]
    print(f"Split data into {len(train_data[0])} training records and {len(val_data[0])} validation records.")
    return train_data, val_data



"""
Filename: train.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Merge dataset.h5 files.
"""

import h5py
import numpy as np
import sys, os
import time
# from utils import *
import argparse
import openspliceai.create_data.utils as utils

donor_motif_counts = {}  # Initialize counts
acceptor_motif_counts = {}  # Initialize counts

def merge_dataset(args):
    print("--- Merging dataset.h5 ... ---")
    start_time = time.time()

    dataset_ls = [] 
    if args.chr_split == 'test':
        dataset_ls.append('test')
    elif args.chr_split == 'train-test':
        dataset_ls.append('test')
        dataset_ls.append('train')
    for dataset_type in dataset_ls:
        print(("\tProcessing %s ..." % dataset_type))
        # if args.biotype =="non-coding":
        #     input_file = f"{args.output_dir}/datafile_{dataset_type}_ncRNA.h5"
        #     output_file = f"{args.output_dir}/dataset_{dataset_type}_ncRNA.h5"
        # elif args.biotype =="protein-coding":
        #     input_file = f"{args.output_dir}/datafile_{dataset_type}.h5"
        #     output_file = f"{args.output_dir}/dataset_{dataset_type}.h5"

        os.makedirs(args.output_dir, exist_ok=True)
        output_file = f"{args.output_dir}/dataset_{dataset_type}.h5"
        h5f2 = h5py.File(output_file, 'w')
        print(f"\t output {output_file} ... ")

        key_counter = 0

        for input_dir in args.input_dir:
            input_file = f"{input_dir}/dataset_{dataset_type}.h5"
            print(f"\tReading {input_file} ... ")
            with h5py.File(input_file, 'r') as h5f:
                x_keys = [key for key in h5f.keys() if key.startswith('X')]
                y_keys = [key for key in h5f.keys() if key.startswith('Y')]
                x_keys.sort(key=lambda x: int(x[1:]))  # Sort by numeric value
                y_keys.sort(key=lambda y: int(y[1:]))  # Sort by numeric value
                for x_key, y_key in zip(x_keys, y_keys):
                    new_x_key = f'X{key_counter}'
                    new_y_key = f'Y{key_counter}'
                    h5f2.create_dataset(new_x_key, data=h5f[x_key][:])
                    h5f2.create_dataset(new_y_key, data=h5f[y_key][:])
                    key_counter += 1
        h5f2.close()
    print(f"--- {time.time() - start_time} seconds ---")





        # CHUNK_SIZE = 100
        # seq_num = SEQ.shape[0]
        # print("seq_num: ", seq_num)
        # print("STRAND.shape[0]: ", STRAND.shape[0])
        # print("TX_START.shape[0]: ", TX_START.shape[0])
        # print("TX_END.shape[0]: ", TX_END.shape[0])
        # print("LABEL.shape[0]: ", LABEL.shape[0])
        # # # Check motif
        # # for idx in range(seq_num):
        # #     label_decode = LABEL[idx].decode('ascii')
        # #     seq_decode = SEQ[idx].decode('ascii')
        # #     strand_decode = STRAND[idx].decode('ascii')
        # #     label_int = [int(char) for char in label_decode]
        # #     utils.check_and_count_motifs(seq_decode, label_int, strand_decode, donor_motif_counts, acceptor_motif_counts)
        # # utils.print_motif_counts(donor_motif_counts, acceptor_motif_counts)
        # # Create dataset
        # for i in range(seq_num//CHUNK_SIZE):
        #     # Each dataset has CHUNK_SIZE genes
        #     if (i+1) == seq_num//CHUNK_SIZE:
        #         NEW_CHUNK_SIZE = CHUNK_SIZE + seq_num%CHUNK_SIZE
        #     else:
        #         NEW_CHUNK_SIZE = CHUNK_SIZE
        #     X_batch = []
        #     Y_batch = [[] for t in range(1)]
        #     for j in range(NEW_CHUNK_SIZE):
        #         idx = i*CHUNK_SIZE + j
        #         seq_decode = SEQ[idx].decode('ascii')
        #         strand_decode = STRAND[idx].decode('ascii')
        #         tx_start_decode = TX_START[idx].decode('ascii')
        #         tx_end_decode = TX_END[idx].decode('ascii')
        #         label_decode = LABEL[idx].decode('ascii')
        #         fixed_seq = utils.replace_non_acgt_to_n(seq_decode)
        #         X, Y = utils.create_datapoints(fixed_seq, strand_decode, label_decode)                
        #         X_batch.extend(X)
        #         for t in range(1):
        #             Y_batch[t].extend(Y[t])
        #     X_batch = np.asarray(X_batch).astype('int8')
        #     print("X_batch.shape: ", X_batch.shape)
        #     for t in range(1):
        #         Y_batch[t] = np.asarray(Y_batch[t]).astype('int8')
        #     print("len(Y_batch[0]): ", len(Y_batch[0]))
        #     h5f2.create_dataset('X' + str(i), data=X_batch)
        #     h5f2.create_dataset('Y' + str(i), data=Y_batch)
        # h5f2.close()
    print("--- %s seconds ---" % (time.time() - start_time))


"""
Filename: paralogs.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Remove paralogous sequences between train and test datasets using mappy.
"""

import mappy as mp
import numpy as np
import logging
import h5py
import tempfile
import os

def remove_paralogous_sequences(train_data, test_data, min_identity, min_coverage, output_dir, exp):
    """
    Remove paralogous sequences between train and test datasets using mappy.
    
    :param train_data: List of lists containing train data (NAME, CHROM, STRAND, TX_START, TX_END, SEQ, LABEL)
    :param test_data: List of lists containing test data (NAME, CHROM, STRAND, TX_START, TX_END, SEQ, LABEL)
    :param min_identity: Minimum identity for sequences to be considered paralogous
    :param min_coverage: Minimum coverage for sequences to be considered paralogous
    :return: Tuple of (filtered_train_data, filtered_test_data)
    """
    print(f"\tStarting paralogy removal process...")
    print(f"\tInitial train set size: {len(train_data[0])}")
    print(f"\tInitial test set size: {len(test_data[0])}")
    train_seqs = train_data[5]  # SEQ is at index 5
    # Create a temporary file with training sequences
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as temp_file:
        for i, seq in enumerate(train_seqs):
            temp_file.write(f">seq{i}\n{seq}\n")
        temp_filename = temp_file.name
    # Create a mappy index from the training sequences
    print("\tCreating mappy index from training sequences...")
    try:
        aligner = mp.Aligner(temp_filename, preset="map-ont")
        if not aligner:
            raise Exception("\tFailed to load/build index")
    except Exception as e:
        logging.error(f"\tError creating mappy aligner: {str(e)}")
        os.unlink(temp_filename)
        return train_data, test_data
    filtered_test_data = [[] for _ in range(len(test_data))]
    paralogous_count = 0
    total_count = len(test_data[5])
    fw = open(f"{output_dir}removed_paralogs_{exp}.txt", "w")
    print("\tStarting to process test sequences...")
    for i in range(total_count):  # Iterate over test sequences
        test_seq = test_data[5][i]
        is_paralogous = False
        for hit in aligner.map(test_seq):
            identity = hit.mlen / hit.blen
            coverage = hit.blen / len(test_seq)
            fw.write(f"{test_data[0][i]}\t{identity}\t{coverage}\n")
            if identity >= min_identity and coverage >= min_coverage:
                # fw.write(f"{test_data[0][i]}\t{identity}\t{coverage}\n")
                print(f"\t\tParalogs detected: Identity: {identity}, Coverage: {coverage}")
                is_paralogous = True
                paralogous_count += 1
                break
        if not is_paralogous:
            for j in range(len(test_data)):
                filtered_test_data[j].append(test_data[j][i])
        # Log progress every 1000 sequences
        if (i + 1) % 1000 == 0:
            print(f"\tProcessed {i + 1}/{total_count} sequences...")
    print(f"\tParalogy removal process completed.")
    print(f"\tNumber of paralogous sequences removed: {paralogous_count}")
    print(f"\tFinal {exp} set size: {len(filtered_test_data[0])}")
    print(f"\tPercentage of {exp} set removed: {(paralogous_count / total_count) * 100:.2f}%")
    fw.close()
    return train_data, filtered_test_data


def write_h5_file(output_dir, data_type, data):
    """
    Write the data to an h5 file.
    """
    h5fname = output_dir + f'datafile_{data_type}.h5'
    h5f = h5py.File(h5fname, 'w')
    dt = h5py.string_dtype(encoding='utf-8')
    
    dataset_names = ['NAME', 'CHROM', 'STRAND', 'TX_START', 'TX_END', 'SEQ', 'LABEL']
    for i, name in enumerate(dataset_names):
        h5f.create_dataset(name, data=np.asarray(data[i], dtype=dt), dtype=dt)    
    h5f.close()


'''
variant.py
This command annotates variants in a VCF file using SpliceAI-toolkit. It reads the input VCF file, annotates 
each variant with delta scores and delta positions, and writes the annotated variants to an output VCF file. 
It uses the Annotator class to annotate variants based on the reference genome and annotation provided. The 
annotated variants are written to the output VCF file with the 'SpliceAI' INFO field containing the delta 
scores and delta positions for acceptor gain (AG), acceptor loss (AL), donor gain (DG), and donor loss (DL). 
'''

import logging
import pysam
import numpy as np
from openspliceai.variant.utils import *
from tqdm import tqdm
import os

# NOTE: if running with gpu, note that cudnn version should be 8.9.6 or higher, numpy <2.0.0

def variant(args):
    print("Running SpliceAI-toolkit with 'variant' mode")
    start_time = time.time()
    
    # Set up logging
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

    # Error handling for required arguments
    if None in [args.input_vcf, args.output_vcf, args.ref_genome, args.annotation, args.model, args.flanking_size]:
        logging.error('Usage: openspliceai [-h] [-m [model]] [-f [flanking_size]] [-I [input]] [-O [output]] -R reference -A annotation '
                      '[-D [distance]] [-M [mask]]')
        exit(1)

    # Define arguments
    ref_genome = args.ref_genome
    annotation = args.annotation
    input_vcf = args.input_vcf
    output_vcf = args.output_vcf
    distance = args.distance
    mask = args.mask
    model = args.model
    flanking_size = args.flanking_size
    model_type = args.model_type
    precision = args.precision
    
    print(f'''Running with genome: {ref_genome}, annotation: {annotation}, 
          model(s): {model}, model_type: {model_type}, 
          input: {input_vcf}, output: {output_vcf}, 
          distance: {distance}, mask: {mask}, flanking_size: {flanking_size}, precision: {precision}''')

    # Reading input VCF file
    print('\t[INFO] Reading input VCF file')
    try:
        vcf = pysam.VariantFile(input_vcf)
    except (IOError, ValueError) as e:
        logging.error('Error reading input file: {}'.format(e))
        exit(1)

    # Adding annotation to the header
    header = vcf.header
    header.add_line('##INFO=<ID=OpenSpliceAI,Number=.,Type=String,Description="OpenSpliceAI variant '
                    'annotation. These include delta scores (DS) and delta positions (DP) for '
                    'acceptor gain (AG), acceptor loss (AL), donor gain (DG), and donor loss (DL). '
                    'Format: ALLELE|SYMBOL|DS_AG|DS_AL|DS_DG|DS_DL|DP_AG|DP_AL|DP_DG|DP_DL">')

    # Generating output VCF file
    print('\t[INFO] Generating output VCF file')
    os.makedirs(os.path.dirname(output_vcf), exist_ok=True)
    try:
        output = pysam.VariantFile(output_vcf, mode='w', header=header)
    except (IOError, ValueError) as e:
        logging.error('Error generating output VCF file: {}'.format(e))
        exit(1)

    # Setup the Annotator based on reference genome and annotation
    logging.info('Initializing Annotator class')
    ann = Annotator(ref_genome, annotation, model, model_type, flanking_size)

    # Obtain delta score for each variant in VCF
    for record in tqdm(vcf):
        scores = get_delta_scores(record, ann, distance, mask, flanking_size, precision)
        if scores:
            record.info['OpenSpliceAI'] = scores
        output.write(record)

    # Close input and output VCF files
    vcf.close()
    output.close()
    logging.info('Annotation completed and written to output VCF file')
    
    print("--- %s seconds ---" % (time.time() - start_time))


"""
Filename: openspliceai.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Main script to run OpenSpliceAI toolkit.
"""

import argparse
import sys
from openspliceai import header
from openspliceai.create_data import create_datafile, create_dataset, verify_h5_file
from openspliceai.train import train
# from openspliceai.test import test
from openspliceai.calibrate import calibrate
from openspliceai.transfer import transfer
from openspliceai.predict import predict
from openspliceai.variant import variant

__VERSION__ = header.__version__

def parse_args_create_data(subparsers):
    parser_create_data = subparsers.add_parser('create-data', help='Create dataset for your genome for SpliceAI model training')
    parser_create_data.add_argument('--annotation-gff', type=str, required=True, help='Path to the GFF file')
    parser_create_data.add_argument('--genome-fasta', type=str, required=True, help='Path to the FASTA file')
    parser_create_data.add_argument('--output-dir', type=str, required=True, help='Output directory to save the data')
    parser_create_data.add_argument('--parse-type', type=str, default='canonical', choices=['canonical', 'all_isoforms'], help='Type of transcript processing')
    parser_create_data.add_argument('--biotype', type=str, default='protein-coding', choices=['protein-coding', 'non-coding', 'all'], help='Biotype of transcript processing')
    parser_create_data.add_argument('--chr-split', type=str, choices=['train-test','test'], default='train-test', help='Whether to obtain testing or both training and testing groups')
    parser_create_data.add_argument('--split-method', type=str, choices=['random', 'human'], default='random', help='Chromosome split method for training and testing dataset')
    parser_create_data.add_argument('--val_split_ratio', type=float, default=0.1, help='Ratio of training and validation dataset')
    parser_create_data.add_argument('--split-ratio', type=float, default=0.8, help='Ratio of training and testing dataset')
    parser_create_data.add_argument('--canonical-only', action='store_true', default=False, help='Flag to obtain only canonical splice site pairs')
    parser_create_data.add_argument('--flanking-size', type=int, default=80, help='Sum of flanking sequence lengths on each side of input (i.e. 40+40)')
    parser_create_data.add_argument('--verify-h5', action='store_true', default=False, help='Verify the generated HDF5 file(s)')
    parser_create_data.add_argument('--remove-paralogs', action='store_true', default=False, help='Remove paralogous sequences between training and testing dataset')
    parser_create_data.add_argument('--min-identity', type=float, default=0.8, help='Minimum minimap2 alignment identity for paralog removal between training and testing dataset')
    parser_create_data.add_argument('--min-coverage', type=float, default=0.5, help='Minimum minimap2 alignment coverage for paralog removal between training and testing dataset')
    parser_create_data.add_argument('--write-fasta', action='store_true', default=False, help='Flag to write out sequences into fasta files')


def parse_args_train(subparsers):
    parser_train = subparsers.add_parser('train', help='Train the SpliceAI model')
    parser_train.add_argument('--epochs', '-n', type=int, default=10, help='Number of epochs for training')
    parser_train.add_argument('--scheduler', '-s', type=str, default="MultiStepLR", choices=["MultiStepLR", "CosineAnnealingWarmRestarts"], help="Learning rate scheduler")
    parser_train.add_argument('--early-stopping', '-E', action='store_true', default=False, help='Enable early stopping')
    parser_train.add_argument("--patience", '-P', type=int, default=2, help="Number of epochs to wait before early stopping")
    parser_train.add_argument('--output-dir', '-o', type=str, required=True, help='Output directory to save the data')
    parser_train.add_argument('--project-name', '-p', type=str, required=True, help="Project name for the train experiment")
    parser_train.add_argument('--exp-num', '-e', type=str, default="0", help="Experiment number")
    parser_train.add_argument('--flanking-size', '-f', type=int, default=80, choices=[80, 400, 2000, 10000], help="Flanking sequence size")
    parser_train.add_argument('--random-seed', '-r', type=int, default=42, help="Random seed for reproducibility")
    parser_train.add_argument('--train-dataset', '-train', type=str, required=True, help="Path to the training dataset")
    parser_train.add_argument('--test-dataset', '-test', type=str, required=True, help="Path to the testing dataset")
    parser_train.add_argument("--loss", '-l', type=str, default='cross_entropy_loss', choices=["cross_entropy_loss", "focal_loss"], help="Loss function for training")
    parser_train.add_argument('--model', '-m', default="SpliceAI", type=str)


def parse_args_test(subparsers):
    parser_test = subparsers.add_parser('test', help='Test the SpliceAI model')
    parser_test.add_argument("--pretrained-model", '-m', type=str, required=True, help="Path to the pre-trained model")
    parser_test.add_argument('--output-dir', '-o', type=str, required=True, help='Output directory to save the data')
    parser_test.add_argument('--project-name', '-p', type=str, required=True, help="Project name for the fine-tuning experiment")
    parser_test.add_argument('--exp-num', '-e', type=str, default=0, help="Experiment number")
    parser_test.add_argument('--flanking-size', '-f', type=int, default=80, choices=[80, 400, 2000, 10000], help="Flanking sequence size")
    parser_test.add_argument('--random-seed', '-r', type=int, default=42, help="Random seed for reproducibility")
    parser_test.add_argument('--test-dataset', '-test', type=str, required=True, help="Path to the testing dataset")
    parser_test.add_argument("--loss", '-l', type=str, default='cross_entropy_loss', choices=["cross_entropy_loss", "focal_loss"], help="Loss function for training")
    parser_test.add_argument('--test-target', '-t', default="OpenSpliceAI", choices=["OpenSpliceAI", "SpliceAI-Keras"], type=str)
    parser_test.add_argument('--log-dir', '-L', default="TEST_LOG", type=str)


def parse_args_calibrate(subparsers):
    parser_calibrate = subparsers.add_parser('calibrate', help='Calibrate the SpliceAI model')
    parser_calibrate.add_argument('--epochs', '-n', type=int, default=10, help='Number of epochs for training')
    parser_calibrate.add_argument('--early-stopping', '-E', action='store_true', default=False, help='Enable early stopping')
    parser_calibrate.add_argument("--patience", '-P', type=int, default=2, help="Number of epochs to wait before early stopping")
    parser_calibrate.add_argument("--output-dir", '-o', type=str, required=True, help="Output directory for model checkpoints and logs")
    parser_calibrate.add_argument("--project-name", '-p', type=str, required=True, help="Project name for the fine-tuning experiment")
    parser_calibrate.add_argument("--exp-num", '-e', type=int, default=0, help="Experiment number")
    parser_calibrate.add_argument("--flanking-size", '-f', type=int, default=80, choices=[80, 400, 2000, 10000], help="Flanking sequence size")
    parser_calibrate.add_argument("--random-seed", '-r', type=int, default=42, help="Random seed for reproducibility")
    parser_calibrate.add_argument("--temperature-file", '-T', type=str, default=None, required=False, help="Path to the temperature file")
    parser_calibrate.add_argument("--pretrained-model", '-m', type=str, required=True, help="Path to the pre-trained model")
    parser_calibrate.add_argument("--train-dataset", '-train', type=str, required=True, help="Path to the training dataset")
    parser_calibrate.add_argument("--test-dataset", '-test', type=str, required=True, help="Path to the testing dataset")
    parser_calibrate.add_argument("--loss", '-l', type=str, default='cross_entropy_loss', choices=["cross_entropy_loss", "focal_loss"], help="Loss function for fine-tuning")


def parse_args_transfer(subparsers):
    parser_transfer = subparsers.add_parser('transfer', help='transfer a pre-trained SpliceAI model on new data.')
    parser_transfer.add_argument('--epochs', '-n', type=int, default=10, help='Number of epochs for training')
    parser_transfer.add_argument('--scheduler', '-s', type=str, default="MultiStepLR", choices=["MultiStepLR", "CosineAnnealingWarmRestarts"], help="Learning rate scheduler")
    parser_transfer.add_argument('--early-stopping', '-E', action='store_true', default=False, help='Enable early stopping')
    parser_transfer.add_argument("--patience", '-P', type=int, default=2, help="Number of epochs to wait before early stopping")
    parser_transfer.add_argument("--output-dir", '-o', type=str, required=True, help="Output directory for model checkpoints and logs")
    parser_transfer.add_argument("--project-name", '-p', type=str, required=True, help="Project name for the fine-tuning experiment")
    parser_transfer.add_argument("--exp-num", '-e', type=int, default=0, help="Experiment number")
    parser_transfer.add_argument("--flanking-size", '-f', type=int, default=80, choices=[80, 400, 2000, 10000], help="Flanking sequence size")
    parser_transfer.add_argument("--random-seed", '-r', type=int, default=42, help="Random seed for reproducibility")
    parser_transfer.add_argument("--pretrained-model", '-m', type=str, required=True, help="Path to the pre-trained model")
    parser_transfer.add_argument("--train-dataset", '-train', type=str, required=True, help="Path to the training dataset")
    parser_transfer.add_argument("--test-dataset", '-test', type=str, required=True, help="Path to the testing dataset")
    parser_transfer.add_argument("--loss", '-l', type=str, default='cross_entropy_loss', choices=["cross_entropy_loss", "focal_loss"], help="Loss function for fine-tuning")
    parser_transfer.add_argument("--unfreeze-all", '-A', action='store_true', default=True, help='Unfreeze all layers for fine-tuning')
    parser_transfer.add_argument("--unfreeze", '-u', type=int, default=1, help="Number of layers to unfreeze for fine-tuning")


def parse_args_predict(subparsers):
    parser_predict = subparsers.add_parser('predict', help='Predict splice sites in a given sequence using the SpliceAI model')
    parser_predict.add_argument('--input-sequence', '-i', type=str, required=True, help="Path to FASTA file of the input sequence")
    parser_predict.add_argument('--model', '-m', type=str, required=True, help='Path to a PyTorch SpliceAI model file')
    parser_predict.add_argument('--flanking-size', '-f', type=int, required=True, help='Sum of flanking sequence lengths on each side of input (i.e. 40+40)')
    parser_predict.add_argument('--output-dir', '-o', type=str, default="./predict_out", help='Output directory to save the data')
    parser_predict.add_argument('--annotation-file', '-a', type=str, required=False, help="Path to GFF file of coordinates for genes")
    parser_predict.add_argument('--threshold', '-t', type=float, default=1e-6, help="Threshold to determine acceptor and donor sites")
    parser_predict.add_argument('--predict-all', '-p', action='store_true', required=False, help="Writes all collected predictions to an intermediate file (Warning: on full genomes, will consume much space.)")
    parser_predict.add_argument('--debug', '-D', action='store_true', required=False, help="Run in debug mode (debug statements are printed to stderr)")
    '''AM: very optional flags below vv'''
    parser_predict.add_argument('--hdf-threshold', type=int, default=0, help='Maximum size before reading sequence into an HDF file for storage')
    parser_predict.add_argument('--flush-threshold', type=int, default=500, help='Maximum number of predictions before flushing to file')
    parser_predict.add_argument('--split-threshold', type=int, default=1500000, help='Maximum length of FASTA entry before splitting')
    parser_predict.add_argument('--chunk-size', type=int, default=100, help='Chunk size for loading HDF5 dataset')


def parse_args_variant(subparsers):
    parser_variant = subparsers.add_parser('variant', help='Label genetic variations with their predicted effects on splicing.')
    parser_variant.add_argument('-R', '--ref-genome', metavar='reference', required=True, help='path to the reference genome fasta file')
    parser_variant.add_argument('-A', '--annotation', metavar='annotation', required=True, help='"grch37" (GENCODE V24lift37 canonical annotation file in '
                                                                                'package), "grch38" (GENCODE V24 canonical annotation file in '
                                                                                'package), or path to a similar custom gene annotation file')
    parser_variant.add_argument('-I', '--input-vcf', metavar='input', nargs='?', default=sys.stdin, help='path to the input VCF file, defaults to standard in')
    parser_variant.add_argument('-O', '--output-vcf', metavar='output', nargs='?', default=sys.stdout, help='path to the output VCF file, defaults to standard out')
    parser_variant.add_argument('-D', '--distance', metavar='distance', nargs='?', default=50, type=int, choices=range(0, 5000),
                                    help='maximum distance between the variant and gained/lost splice '
                                        'site, defaults to 50')
    parser_variant.add_argument('-M', '--mask', metavar='mask', nargs='?', default=0, type=int, choices=[0, 1], 
                                    help='mask scores representing annotated acceptor/donor gain and '
                                        'unannotated acceptor/donor loss, defaults to 0')
    '''AM: newly added flags below vv'''
    parser_variant.add_argument('--model', '-m', default="SpliceAI", type=str, help='Path to a SpliceAI model file, or path to a directory of SpliceAI models, or "SpliceAI" for the default model')
    parser_variant.add_argument('--flanking-size', '-f', type=int, default=80, help='Sum of flanking sequence lengths on each side of input (i.e. 40+40)')
    parser_variant.add_argument('--model-type', '-t', type=str, choices=['keras', 'pytorch'], default='pytorch', help='Type of model file (keras or pytorch)')
    parser_variant.add_argument('--precision', '-p', type=int, default=2, help='Number of decimal places to round the output scores')
 

def parse_args(arglist):
    parser = argparse.ArgumentParser(description='OpenSpliceAI toolkit to help you retrain your own splice site predictor')
    # Create a parent subparser to house the common subcommands.
    subparsers = parser.add_subparsers(dest='command', required=True, help='Subcommands: create-data, train, calibrate, predict, transfer, variant')
    parse_args_create_data(subparsers)
    parse_args_train(subparsers)
    # parse_args_test(subparsers)
    parse_args_calibrate(subparsers)
    parse_args_transfer(subparsers)
    parse_args_predict(subparsers)
    parse_args_variant(subparsers)
    if arglist is not None:
        args = parser.parse_args(arglist)
    else:
        args = parser.parse_args()
    return args


def main(arglist=None):
    # ANSI Shadow
    banner = '''
============================================================
Deep learning framework that decodes splicing across species
============================================================


 ██████╗ ██████╗ ███████╗███╗   ██╗███████╗██████╗ ██╗     ██╗ ██████╗███████╗ █████╗ ██╗
██╔═══██╗██╔══██╗██╔════╝████╗  ██║██╔════╝██╔══██╗██║     ██║██╔════╝██╔════╝██╔══██╗██║
██║   ██║██████╔╝█████╗  ██╔██╗ ██║███████╗██████╔╝██║     ██║██║     █████╗  ███████║██║
██║   ██║██╔═══╝ ██╔══╝  ██║╚██╗██║╚════██║██╔═══╝ ██║     ██║██║     ██╔══╝  ██╔══██║██║
╚██████╔╝██║     ███████╗██║ ╚████║███████║██║     ███████╗██║╚██████╗███████╗██║  ██║██║
 ╚═════╝ ╚═╝     ╚══════╝╚═╝  ╚═══╝╚══════╝╚═╝     ╚══════╝╚═╝ ╚═════╝╚══════╝╚═╝  ╚═╝╚═╝
    '''
    print(banner, file=sys.stderr)
    print(f"{__VERSION__}\n", file=sys.stderr)
    args = parse_args(arglist)
    
    if args.command == 'create-data':
        create_datafile.create_datafile(args)
        create_dataset.create_dataset(args)
        if args.verify_h5:
            verify_h5_file.verify_h5(args)
    elif args.command == 'train':
        train.train(args)
    # elif args.command == 'test':
    #     test.test(args)
    elif args.command == 'calibrate':
        calibrate.calibrate(args)
    elif args.command == 'transfer':
        transfer.transfer(args)
    elif args.command == 'predict':
        predict.predict_cli(args)
    elif args.command == 'variant':
        variant.variant(args)


"""
Filename: transfer.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Transfer-learning for OpenSpliceAI model.
"""

import numpy as np
import torch
import torch.optim as optim
from openspliceai.train_base.openspliceai import *
from openspliceai.train_base.utils import *
from openspliceai.constants import *
from sklearn.metrics import precision_recall_fscore_support, accuracy_score


def initialize_model_and_optim_transfer(device, flanking_size, epochs, scheduler,
                               pretrained_model, unfreeze, unfreeze_all):
    L = 32
    N_GPUS = 2
    W = np.asarray([11, 11, 11, 11])
    AR = np.asarray([1, 1, 1, 1])
    BATCH_SIZE = 18 * N_GPUS
    if int(flanking_size) == 80:
        W = np.asarray([11, 11, 11, 11])
        AR = np.asarray([1, 1, 1, 1])
        BATCH_SIZE = 18 * N_GPUS
    elif int(flanking_size) == 400:
        W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11])
        AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4])
        BATCH_SIZE = 18 * N_GPUS
    elif int(flanking_size) == 2000:
        W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11,
                        21, 21, 21, 21])
        AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4,
                        10, 10, 10, 10])
        BATCH_SIZE = 12 * N_GPUS
    elif int(flanking_size) == 10000:
        W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11,
                        21, 21, 21, 21, 41, 41, 41, 41])
        AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4,
                        10, 10, 10, 10, 25, 25, 25, 25])
        BATCH_SIZE = 6 * N_GPUS    
    CL = 2 * np.sum(AR * (W - 1))
    print("\033[1mContext nucleotides: %d\033[0m" % (CL))
    print("\033[1mSequence length (output): %d\033[0m" % (SL))
    # Initialize the model
    model = SpliceAI(L, W, AR).to(device)
    # # Print the shapes of the parameters in the initialized model
    # print("\nInitialized model parameter shapes:")
    # for name, param in model.named_parameters():
    #     print(f"{name}: {param.shape}", end=", ")

    # Load the pretrained model
    state_dict = torch.load(pretrained_model, map_location=device)

    # Filter out unnecessary keys and load matching keys into model
    model_dict = model.state_dict()
    state_dict = {k: v for k, v in state_dict.items() if k in model_dict and v.size() == model_dict[k].size()}

    # Load state dict into the model
    missing_keys, unexpected_keys = model.load_state_dict(state_dict, strict=False)

    # Print missing and unexpected keys
    print("\nMissing keys:", missing_keys)
    print("Unexpected keys:", unexpected_keys)

    print("\n unfreeze_all:", unfreeze_all)
    if not unfreeze_all:
        # Freeze all layers first
        for param in model.parameters():
            param.requires_grad = False
        # Unfreeze the last `unfreeze` layers
        if unfreeze > 0:
            # Unfreeze the last few layers (example: last residual unit)
            for param in model.residual_units[-unfreeze].parameters():
                param.requires_grad = True
    # Set up optimizer and scheduler
    optimizer = optim.AdamW(filter(lambda p: p.requires_grad, model.parameters()), lr=1e-4)
    if scheduler == "MultiStepLR":
        scheduler_obj = optim.lr_scheduler.MultiStepLR(optimizer, milestones=[epochs-4, epochs-3, epochs-2, epochs-1], gamma=0.5)
    elif scheduler == "CosineAnnealingWarmRestarts":
        scheduler_obj = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(
            optimizer, T_0=5, T_mult=1, eta_min=1e-5, last_epoch=-1)    
    params = {'L': L, 'W': W, 'AR': AR, 'CL': CL, 'SL': SL, 'BATCH_SIZE': BATCH_SIZE, 'N_GPUS': N_GPUS}
    return model, optimizer, scheduler_obj, params


def transfer(args):
    print('Running OpenSpliceAI with transfer mode.')
    # assert training_target in ["RefSeq", "MANE", "SpliceAI", "SpliceAI27"]
    device = setup_environment(args)
    model_output_base, log_output_train_base, log_output_val_base, log_output_test_base = initialize_paths(args)
    train_h5f, valid_h5f, test_h5f, batch_num = load_datasets(args)
    train_idxs, val_idxs, test_idxs = generate_indices(train_h5f, valid_h5f, test_h5f)
    model, optimizer, scheduler, params = initialize_model_and_optim_transfer(device, args.flanking_size, args.epochs, args.scheduler, args.pretrained_model, args.unfreeze, args.unfreeze_all)
    
    params["RANDOM_SEED"] = args.random_seed
    train_metric_files = create_metric_files(log_output_train_base)
    valid_metric_files = create_metric_files(log_output_val_base)
    test_metric_files = create_metric_files(log_output_test_base)
    train_model(model, optimizer, scheduler, train_h5f, valid_h5f, test_h5f, train_idxs, 
                val_idxs, test_idxs, model_output_base, args, device, params, train_metric_files, valid_metric_files, test_metric_files)
    train_h5f.close()
    valid_h5f.close()
    test_h5f.close()


"""
Filename: verify_h5_file.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Functions to process sequences to/from .h5 datasets.
"""

import h5py
import torch
import matplotlib.pyplot as plt
import time 
import h5py

def verify_h5(args):
    """
    Verifies the integrity of the created testing and/or training h5 datasets.
    """
    # record start time for benchmark
    print("--- Step 3: Verifying integrity of h5 file ... ---")
    start_time = time.time()

    # construct the filename and open h5 file
    dataset_ls = [] 
    if args.chr_split == 'test':
        dataset_ls.append('test')
    elif args.chr_split == 'train-test':
        dataset_ls.append('test')
        dataset_ls.append('train')
    for dataset_type in dataset_ls:
        if args.biotype =="non-coding":
            filename = f"{args.output_dir}/dataset_{dataset_type}_ncRNA.h5"
            figname = f"{args.output_dir}/verify_{dataset_type}_ncRNA.png"
        elif args.biotype =="protein-coding":
            filename = f"{args.output_dir}/dataset_{dataset_type}.h5"
            figname = f"{args.output_dir}/verify_{dataset_type}.png"

        print(f'Verifying {filename}...')
        with h5py.File(filename, 'r') as hf:
            # print the available dataset keys in the file
            print(f"Dataset keys: {list(hf.keys())}\n\n")

            # convert datasets to PyTorch tensors and display their shapes
            X0_tensor = torch.from_numpy(hf['X0'][:]).float()  
            Y0_tensor = torch.from_numpy(hf['Y0'][:]) 
            print(f"X0 shape: {X0_tensor.shape}, Y0 shape: {Y0_tensor.shape}")

            # process a specific dataset ('X3') for visualization
            x = torch.from_numpy(hf['X3'][:]).float()  
            y = torch.from_numpy(hf['Y3'][:])  
            print(f"x[0].shape: {x[0].shape}, y[0].shape: {y[0].shape}")

            # plot the sum of the last entry in the 'X3' dataset along its rows
            fig = plt.figure(figsize=(7, 3))
            ax = fig.add_subplot(111)
            ax.set_xlim([0, 15000])
            ax.plot(x[-1].sum(axis=1))  # Sum over rows and plot
            fig.savefig(figname, dpi=300, bbox_inches='tight')
            plt.close(fig)

            # additional info
            print("x[0].sum(axis=1): ", len(x[len(x)-1].sum(axis=1))) # Length of row sum in the last 'X3' entry
            print("x[0].sum(axis=0): ", len(x[len(x)-1].sum(axis=0))) # Length of column sum in the last 'X3' entry
            print(f"(x[0].sum(axis=1) == 0).sum(): {(x[0].sum(axis=1) == 0).sum()}") # Number of zero-sum rows in the first 'X3' entry

    # END
    print(("--- %s seconds ---" % (time.time() - start_time)))


"""
Filename: calibrate.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Calibrate the OpenSpliceAI model.
"""

import os
import time
import torch
from torch.nn import functional as F
from openspliceai.train_base.utils import *
from openspliceai.calibrate.calibrate_utils import *
from openspliceai.calibrate.visualization import *
from openspliceai.calibrate.temperature_scaling import *
from openspliceai.calibrate.model_utils import initialize_model_and_optim

def get_logits_labels(model, loader, device, params):
    """
    Same as before.
    """
    model.eval()
    logits_list, labels_list = [], []
    with torch.no_grad():
        for inputs, labels in tqdm(loader, desc='Collecting logits and labels'):
            inputs, labels = inputs.to(device), labels.to(device)
            inputs, labels = clip_datapoints(inputs, labels, params["CL"], CL_max, params["N_GPUS"])
            logits = model(inputs)
            logits_list.append(logits)
            labels_list.append(labels)

    logits = torch.cat(logits_list).permute(0, 2, 1).contiguous()
    labels = torch.cat(labels_list).permute(0, 2, 1).contiguous().argmax(dim=-1)

    # Flatten logits and labels
    N, L, C = logits.shape
    logits = logits.view(-1, C)
    labels = labels.view(-1).long()
    return logits, labels


def evaluate_and_visualize(calibrated_model, data_loader, device, output_base_dir, dataset_name, params, flanking_size):
    print(f"\n--- Evaluating on {dataset_name} set ---")
    
    results_dir = os.path.join(output_base_dir, "results", dataset_name)
    os.makedirs(results_dir, exist_ok=True)
    calib_data_dir = os.path.join(results_dir, "calibration_data")
    os.makedirs(calib_data_dir, exist_ok=True)
    plots_dir = os.path.join(results_dir, "plots")
    os.makedirs(plots_dir, exist_ok=True)
    
    # Compute logits and labels using the base model
    base_model = calibrated_model.model
    logits, labels = get_logits_labels(base_model, data_loader, device, params)

    # Apply temperature scaling
    logits_scaled = calibrated_model.temperature_scale(logits)

    # Compute metrics
    metric_original_file = os.path.join(results_dir, "metrics_original.txt")
    metric_calibrated_file = os.path.join(results_dir, "metrics_calibrated.txt")
    
    original_nll, original_ece = calibrated_model.compute_ece_nll(logits, labels)
    with open(metric_original_file, 'w') as f:
        f.write(f"Original_NLL\tOriginal_ECE\n")
        f.write(f"{original_nll:.8f}\t{original_ece:.8f}\n")
    print(f"Original NLL: {original_nll:.8f}")
    print(f"Original ECE: {original_ece:.8f}")

    calibrated_nll, calibrated_ece = calibrated_model.compute_ece_nll(logits_scaled, labels)
    with open(metric_calibrated_file, 'w') as f:
        f.write(f"Calibrated_NLL\tCalibrated_ECE\n")
        f.write(f"{calibrated_nll:.8f}\t{calibrated_ece:.8f}\n")
    print(f"Calibrated NLL: {calibrated_nll:.8f}")
    print(f"Calibrated ECE: {calibrated_ece:.8f}")

    # Convert logits to probabilities
    probs = F.softmax(logits, dim=1).detach().cpu().numpy()
    probs_scaled = F.softmax(logits_scaled, dim=1).detach().cpu().numpy()
    labels = labels.cpu().numpy()

    # Plotting, calibration curve computations, etc. remain the same
    for idx in [0, 1, 2]:
        plot_score_distribution(probs, probs_scaled, labels, plots_dir, idx)
    
    classes = ["Non-splice site", "Acceptor site", "Donor site"]
    calibration_data = []
    calibration_data_scaled = []
    for i in range(3):
        class_labels = (labels == i).astype(int)
        class_probs = probs[:, i]
        class_probs_scaled = probs_scaled[:, i]
        
        prob_true, prob_pred, bin_counts = compute_calibration_curve(
            class_labels, class_probs, n_bins=30, strategy='uniform')
        prob_true_scaled, prob_pred_scaled, _ = compute_calibration_curve(
            class_labels, class_probs_scaled, n_bins=30, strategy='uniform')
        
        calibration_data.append((prob_true, prob_pred, bin_counts))
        calibration_data_scaled.append((prob_true_scaled, prob_pred_scaled, bin_counts))
        save_calibration_data(calib_data_dir, classes[i], flanking_size, prob_true, prob_pred, bin_counts, 'original')
        save_calibration_data(calib_data_dir, classes[i], flanking_size, prob_true_scaled, prob_pred_scaled, bin_counts, 'calibrated')

    plot_calibration_curves(calibration_data, calibration_data_scaled, classes, plots_dir)
    brier_uncal, brier_cal = calculate_brier_scores(labels, probs, probs_scaled)
    plot_brier_scores(brier_uncal, brier_cal, classes, plots_dir)
    plot_calibration_map(calibrated_model, device, plots_dir)
    print(f"Results for {dataset_name} set saved to: {results_dir}")
    print("===============================================")


def calibrate(args):
    print("Running OpenSpliceAI with 'calibrate' mode")
    start_time = time.time()
    
    # Create the main output directory structure
    base_output_dir = args.output_dir
    os.makedirs(base_output_dir, exist_ok=True)
    calibration_output_dir = os.path.join(base_output_dir, "calibration")
    os.makedirs(calibration_output_dir, exist_ok=True)
    
    # Set up the device, datasets, and indices
    device = setup_environment(args)
    train_h5f, test_h5f, batch_num = load_datasets(args)
    train_idxs, val_idxs, test_idxs = generate_indices(batch_num, args.random_seed, test_h5f)
    
    # Initialize the model
    model, model_params = initialize_model_and_optim(device, args.flanking_size, args.pretrained_model)

    # -----------------------------
    # NEW: Provide num_classes here
    # (for example 3, if you know you have 3 classes)
    # If it's variable, you can determine dynamically from the model output dimension.
    num_classes = 3

    calibrated_model = ModelWithTemperature(model, num_classes=num_classes)
    print("Initialized calibrated model:", calibrated_model)    
    print("Validation indices count:", len(val_idxs))
    print("Test indices count:", len(test_idxs))
    
    # Create data loaders for the validation (calibration) and test sets
    validation_loader = get_validation_loader(train_h5f, val_idxs, model_params["BATCH_SIZE"])
    test_loader = get_validation_loader(test_h5f, test_idxs, model_params["BATCH_SIZE"])
    
    # Load or determine the temperature vector
    if args.temperature_file:
        calibrated_model.load_temperature(args.temperature_file, validation_loader, model_params)
        # Because we have a vector, show the full array
        print(f"Loaded temperature from {args.temperature_file}: {calibrated_model.temperature.data.cpu().numpy()}")
    else:
        # Calibrate
        calibrated_model.set_temperature(validation_loader, model_params)
        temperature_save_path = os.path.join(base_output_dir, "temperature.pt")
        calibrated_model.save_temperature(temperature_save_path)
        print(f"Saved calibrated temperature to {temperature_save_path}")
    
    # Save the temperature vector in text form
    temperature_txt_save_path = os.path.join(base_output_dir, "temperature.txt")
    with open(temperature_txt_save_path, 'w') as f:
        f.write(str(calibrated_model.temperature.data.cpu().numpy()))
    
    # Save the full calibrated model (including the temperature vector)
    model_save_path = os.path.join(base_output_dir, "calibrated_model.pt")
    torch.save(calibrated_model, model_save_path)
    print(f"Calibrated model saved to: {model_save_path}")
    
    # Evaluate and visualize on the validation set
    evaluate_and_visualize(calibrated_model, validation_loader, device, calibration_output_dir, "validation", model_params, args.flanking_size)
    
    # Evaluate and visualize on the test set
    evaluate_and_visualize(calibrated_model, test_loader, device, calibration_output_dir, "test", model_params, args.flanking_size)
    
    end_time = time.time()
    print(f"\nTotal calibration and evaluation time: {end_time - start_time:.2f} seconds")



###############################################################################
'''This code has functions which process the information in the .h5 files
datafile_{}_{}.h5 and convert them into a format usable by Keras.'''
###############################################################################

# PROCESSING BATCH SIZE
from math import ceil
def ceil_div(x, y):
    """
    Calculate the ceiling of a division between two numbers.

    Parameters:
    - x (int): Numerator
    - y (int): Denominator

    Returns:
    - int: The ceiling of the division result.
    """
    return int(ceil(float(x)/y))

# FOR TESTING PURPOSES
import os, sys
import psutil
def log_memory_usage():
    process = psutil.Process(os.getpid())
    print(f"Memory usage: {process.memory_info().rss / (1024 * 1024)} MB", file=sys.stderr)

# SETUP INITIALIZATION   
def initialize_constants(flanking_size, hdf_threshold_len=0, flush_predict_threshold=500, chunk_size=100, split_fasta_threshold=1500000):
    from openspliceai.constants import SL
    
    assert int(flanking_size) in [80, 400, 2000, 10000]
    
    CL_max = flanking_size                              # context length for sequence prediction (flanking size sum)
    HDF_THRESHOLD_LEN = hdf_threshold_len               # maximum size before reading sequence into an HDF file for storage
    FLUSH_PREDICT_THRESHOLD = flush_predict_threshold   # maximum number of predictions before flushing to file
    CHUNK_SIZE = chunk_size                             # chunk size for loading hdf5 dataset
    SPLIT_FASTA_THRESHOLD = split_fasta_threshold       # maximum length of fasta entry before splitting
    
    return {'SL': SL,
            'CL_max': CL_max, 
            'HDF_THRESHOLD_LEN': HDF_THRESHOLD_LEN, 
            'FLUSH_PREDICT_THRESHOLD': FLUSH_PREDICT_THRESHOLD, 
            'CHUNK_SIZE': CHUNK_SIZE, 
            'SPLIT_FASTA_THRESHOLD': SPLIT_FASTA_THRESHOLD}
    
def initialize_paths(output_dir, flanking_size, sequence_length, model_arch='SpliceAI'):
    """Initialize project directories and create them if they don't exist."""
    BASENAME = f"{model_arch}_{sequence_length}_{flanking_size}"
    model_pred_outdir = f"{output_dir}/{BASENAME}/"
    os.makedirs(model_pred_outdir, exist_ok=True)

    return model_pred_outdir


"""
Filename: model_utils.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Calibrate the OpenSpliceAI model.
"""

import torch
import numpy as np
from openspliceai.constants import *
from openspliceai.train_base.openspliceai import SpliceAI

def initialize_model_and_optim(device, flanking_size, pretrained_model):
    L = 32
    N_GPUS = 2
    W = np.asarray([11, 11, 11, 11])
    AR = np.asarray([1, 1, 1, 1])
    BATCH_SIZE = 18 * N_GPUS
    if int(flanking_size) == 80:
        W = np.asarray([11, 11, 11, 11])
        AR = np.asarray([1, 1, 1, 1])
        BATCH_SIZE = 18 * N_GPUS
    elif int(flanking_size) == 400:
        W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11])
        AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4])
        BATCH_SIZE = 18 * N_GPUS
    elif int(flanking_size) == 2000:
        W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11,
                        21, 21, 21, 21])
        AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4,
                        10, 10, 10, 10])
        BATCH_SIZE = 12 * N_GPUS
    elif int(flanking_size) == 10000:
        W = np.asarray([11, 11, 11, 11, 11, 11, 11, 11,
                        21, 21, 21, 21, 41, 41, 41, 41])
        AR = np.asarray([1, 1, 1, 1, 4, 4, 4, 4,
                        10, 10, 10, 10, 25, 25, 25, 25])
        BATCH_SIZE = 6 * N_GPUS    
    CL = 2 * np.sum(AR * (W - 1))
    print("\033[1mContext nucleotides: %d\033[0m" % (CL))
    print("\033[1mSequence length (output): %d\033[0m" % (SL))
    # Initialize the model
    model = SpliceAI(L, W, AR, apply_softmax=False).to(device)
    # Print the shapes of the parameters in the initialized model
    print("\nInitialized model parameter shapes:")
    for name, param in model.named_parameters():
        print(f"{name}: {param.shape}", end=", ")

    # Load the pretrained model
    state_dict = torch.load(pretrained_model, map_location=device)

    # Filter out unnecessary keys and load matching keys into model
    model_dict = model.state_dict()
    state_dict = {k: v for k, v in state_dict.items() if k in model_dict and v.size() == model_dict[k].size()}

    # Load state dict into the model
    missing_keys, unexpected_keys = model.load_state_dict(state_dict, strict=False)

    # Print missing and unexpected keys
    print("\nMissing keys:", missing_keys)
    print("Unexpected keys:", unexpected_keys)

    params = {'L': L, 'W': W, 'AR': AR, 'CL': CL, 'SL': SL, 'BATCH_SIZE': BATCH_SIZE, 'N_GPUS': N_GPUS}
    return model, params




"""
Filename: utils.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Utility functions for training and testing the OpenSpliceAI model.
"""

import h5py
import platform
import sys
import os
import time
import numpy as np
import torch
from torch.utils.data import TensorDataset, DataLoader
from tqdm import tqdm
from sklearn.metrics import average_precision_score, precision_recall_fscore_support, accuracy_score
from openspliceai.constants import *

def setup_environment(args):
    assert int(args.flanking_size) in [80, 400, 2000, 10000]
    device = setup_device()
    print("device: ", device, file=sys.stderr)
    return device


def initialize_test_paths(args):
    log_output_test_base = initialize_test_paths_inner(
        args.output_dir, args.project_name, args.flanking_size, args.exp_num, 
        args.random_seed, args.test_target, args.log_dir
    )
    return log_output_test_base


def initialize_test_paths_inner(output_dir, project_name, flanking_size, exp_num, random_seed, test_target, log_dir):
    MODEL_VERSION = f"SpliceAI_{project_name}_{flanking_size}_{exp_num}_rs{random_seed}"
    model_train_outdir = f"{output_dir}/{MODEL_VERSION}/{exp_num}/"
    log_output_base = f"{model_train_outdir}{log_dir}/"
    log_output_test_base = f"{log_output_base}TEST/"
    # TEST_LOG_SpliceAI-Keras
    for path in [log_output_base, log_output_test_base]:
        os.makedirs(path, exist_ok=True)
    return log_output_test_base


def load_test_datasets(args):
    test_h5f = h5py.File(args.test_dataset, 'r')
    return test_h5f


def generate_test_indices(random_seed, test_h5f):
    np.random.seed(random_seed)
    test_idxs = np.arange(len(test_h5f.keys()) // 2)
    np.random.shuffle(test_idxs)
    return test_idxs


def clip_datapoints_spliceai27(X, Y, CL, N_GPUS):
    rem = X.shape[0]%N_GPUS
    clip = (CL_max-CL)//2
    if rem != 0 and clip != 0:
        return X[:-rem, clip:-clip], [Y[t][:-rem] for t in range(1)]
    elif rem == 0 and clip != 0:
        return X[:, clip:-clip], [Y[t] for t in range(1)]
    elif rem != 0 and clip == 0:
        return X[:-rem], [Y[t][:-rem] for t in range(1)]
    else:
        return X, [Y[t] for t in range(1)]


class MetricsAccumulator:
    def __init__(self, num_classes):
        self.true_classes = []
        self.predicted_classes = []
        self.num_classes = num_classes

    def update(self, y_true, y_pred):
        self.true_classes.extend(y_true)
        self.predicted_classes.extend(y_pred)

    def calculate_overall_metrics(self):
        true_classes = np.array(self.true_classes)
        predicted_classes = np.array(self.predicted_classes)
        true_classes = true_classes.flatten()
        predicted_classes = predicted_classes.flatten()
        print(f"true_classes: {true_classes.shape}, predicted_classes: {predicted_classes.shape}")
        class_accuracies = classwise_accuracy(true_classes, predicted_classes, self.num_classes)
        overall_accuracy = np.mean(class_accuracies)
        precision, recall, f1, _ = precision_recall_fscore_support(true_classes, predicted_classes, average=None)
        return overall_accuracy, precision, recall, f1, class_accuracies


def calculate_batch_metrics(y_true, y_pred, metric_files, accumulator):
    accumulator.update(y_true, y_pred)
    batch_accuracies = classwise_accuracy(y_true, y_pred, accumulator.num_classes)
    batch_overall_accuracy = np.mean(batch_accuracies)
    print(f"Batch Overall Accuracy: {batch_overall_accuracy}")
    

def process_batch(model, X, Y, params):
    Xc, Yc = clip_datapoints_spliceai27(X, Y, params['CL'], 2)
    Yp = model.predict(Xc, batch_size=params['BATCH_SIZE'])
    return Yc[0], Yp


def test_SpliceAI_Keras_model(model, test_h5f, test_idxs, args, params, test_metric_files):
    print(f"\n{'='*60}")
    start_time = time.time()    
    print("--------------------------------------------------------------")
    print("\n\033[1mValidation set metrics:\033[0m")
    
    np.random.seed(params["RANDOM_SEED"])
    shuffled_idxs = np.random.choice(test_idxs, size=len(test_idxs), replace=False)    
    print("shuffled_idxs: ", shuffled_idxs)
    
    shuffled_idxs = shuffled_idxs  # For testing purposes
    batch_size = 5  # Adjust this based on your memory constraints
    num_classes = 3  # Assuming 3 classes: Non-splice, acceptor, donor
    metrics_accumulator = MetricsAccumulator(num_classes)
    
    # Initialize lists to collect true labels and predicted probabilities
    y_true_acceptor_all = []
    y_pred_acceptor_all = []
    y_true_donor_all = []
    y_pred_donor_all = []

    for i in range(0, len(shuffled_idxs), batch_size):
        batch_idxs = shuffled_idxs[i:i+batch_size]
        batch_ylabel = []
        batch_ypred = []
        
        for idx in batch_idxs:
            X = test_h5f['X' + str(idx)][()]  # Assuming HDF5 dataset; read data
            Y = test_h5f['Y' + str(idx)][()]
            print(f"\nProcessing index {idx}:")
            print(f"\tX.shape: {X.shape}, Y.shape: {Y.shape}")
            Xc, Yc = clip_datapoints_spliceai27(X, Y, params['CL'], 2)
            Yp = model.predict(Xc, batch_size=params['BATCH_SIZE'])
            batch_ylabel.append(Yc[0])  # Yc is a list; take the first element
            batch_ypred.append(Yp)
        batch_ylabel = np.concatenate(batch_ylabel, axis=0)
        batch_ypred = np.concatenate(batch_ypred, axis=0)
        print(f"\nBatch shapes - ylabel: {batch_ylabel.shape}, ypred: {batch_ypred.shape}")
        
        # Reshape to 2D arrays if necessary
        batch_ylabel = batch_ylabel.reshape(-1, num_classes)
        batch_ypred = batch_ypred.reshape(-1, num_classes)
        
        # Extract true labels and predictions for acceptor and donor
        y_true_acceptor = batch_ylabel[:, 1]  # Acceptor class index is 1
        y_pred_acceptor = batch_ypred[:, 1]
        y_true_donor = batch_ylabel[:, 2]     # Donor class index is 2
        y_pred_donor = batch_ypred[:, 2]

        # Accumulate the labels and predictions
        y_true_acceptor_all.append(y_true_acceptor)
        y_pred_acceptor_all.append(y_pred_acceptor)
        y_true_donor_all.append(y_true_donor)
        y_pred_donor_all.append(y_pred_donor)
        
        # Convert probabilities to class predictions
        batch_ypred_classes = np.argmax(batch_ypred, axis=-1)
        batch_ylabel_classes = np.argmax(batch_ylabel, axis=-1)
        calculate_batch_metrics(batch_ylabel_classes, batch_ypred_classes, test_metric_files, metrics_accumulator)
        # Clear variables to free up memory
        del batch_ylabel_classes, batch_ypred_classes

    # Concatenate all batches
    y_true_acceptor_all = np.concatenate(y_true_acceptor_all)
    y_pred_acceptor_all = np.concatenate(y_pred_acceptor_all)
    y_true_donor_all = np.concatenate(y_true_donor_all)
    y_pred_donor_all = np.concatenate(y_pred_donor_all)

    acceptor_topk_accuracy, acceptor_auprc = print_topl_statistics(y_true_donor_all, y_pred_donor_all, test_metric_files["acceptor_topk_all"], ss_type='acceptor', print_top_k=True)
    donor_topk_accuracy, donor_auprc = print_topl_statistics(y_true_donor_all, y_pred_donor_all, test_metric_files["donor_topk_all"], ss_type='donor', print_top_k=True)

    # Calculate and write overall metrics
    overall_accuracy, precision, recall, f1, class_accuracies = metrics_accumulator.calculate_overall_metrics()
    print(f"Overall Accuracy: {overall_accuracy:.4f}")
    for k, v in test_metric_files.items():
        with open(v, 'a') as f:
            if k == "accuracy":
                f.write(f"{overall_accuracy}\n")
            elif k == "donor_topk":
                f.write(f"{donor_topk_accuracy}\n")
            elif k == "donor_auprc":
                f.write(f"{donor_auprc}\n")
            elif k == "acceptor_topk":
                f.write(f"{acceptor_topk_accuracy}\n")
            elif k == "acceptor_auprc":
                f.write(f"{acceptor_auprc}\n")

    ss_types = ["Non-splice", "acceptor", "donor"]
    for i, (acc, prec, rec, f1_score) in enumerate(zip(class_accuracies, precision, recall, f1)):
        print(f"Class {ss_types[i]}:\tAccuracy={acc:.4f}, Precision={prec:.4f}, Recall={rec:.4f}, F1={f1_score:.4f}")
        if ss_types[i] == "Non-splice":
            continue  # Skip writing metrics for Non-splice class
        for k, v in test_metric_files.items():
            with open(v, 'a') as f:
                if k == f"{ss_types[i]}_precision":
                    f.write(f"{prec}\n")
                elif k == f"{ss_types[i]}_recall":
                    f.write(f"{rec}\n")
                elif k == f"{ss_types[i]}_f1":
                    f.write(f"{f1_score}\n")
                elif k == f"{ss_types[i]}_accuracy":
                    f.write(f"{acc}\n")
    print("--------------------------------------------------------------")
    print(f"--- {time.time() - start_time:.2f} seconds ---")
    print("="*60)


def test_model(model, optimizer, test_h5f, test_idxs, args, device, params, test_metric_files):
    print("test_idxs: ", test_idxs)
    print(f"\n{'='*60}")
    start_time = time.time()
    test_loss = valid_epoch(model, test_h5f, test_idxs, params["BATCH_SIZE"], args.loss, device, 
                            params, test_metric_files, args.flanking_size, "test")
    print(f"Testing Loss: {test_loss}")
    print(f"--- {time.time() - start_time:.2f} seconds ---")
    print("="*60)


###########################
# Model training
###########################
def initialize_paths(args):
    model_output_base, log_output_train_base, log_output_val_base, log_output_test_base = initialize_paths_inner(
        args.output_dir, args.project_name, args.flanking_size, args.exp_num, SL, args.loss, args.random_seed
    )
    return model_output_base, log_output_train_base, log_output_val_base, log_output_test_base


def load_datasets(args):
    train_h5f = h5py.File(args.train_dataset, 'r')
    os.path.dirname(args.train_dataset)
    valid_dataset = os.path.join(os.path.dirname(args.train_dataset), os.path.basename(args.train_dataset).replace("train", "validation"))
    valid_h5f = h5py.File(valid_dataset, 'r')
    test_h5f = h5py.File(args.test_dataset, 'r')
    batch_num = len(train_h5f.keys()) // 2
    print("* Batch_num: ", batch_num, file=sys.stderr)
    return train_h5f, valid_h5f, test_h5f, batch_num


def generate_indices(train_h5f, valid_h5f, test_h5f):
    # np.random.seed(random_seed)
    # idxs = np.random.permutation(batch_num)
    # train_idxs = idxs[:int(0.9 * batch_num)]
    # val_idxs = idxs[int(0.9 * batch_num):]
    # test_idxs = np.arange(len(test_h5f.keys()) // 2)
    train_idxs = np.arange(len(train_h5f.keys()) // 2)    
    val_idxs = np.arange(len(valid_h5f.keys()) // 2)        
    test_idxs = np.arange(len(test_h5f.keys()) // 2)
    # # Other approach of splitting test set
    # # Generate and shuffle indices for training set
    # train_idxs = np.arange(batch_num)
    # np.random.shuffle(train_idxs)    
    # # Generate indices for test set
    # test_idxs = np.arange(len(test_h5f.keys()) // 2)
    # np.random.shuffle(test_idxs)
    # # Split test set into test and validation
    # val_size = int(0.5 * len(test_idxs))  # 10% for validation
    # val_idxs = test_idxs[:val_size]
    # test_idxs = test_idxs[val_size:]
    return train_idxs, val_idxs, test_idxs


def create_metric_files(log_output_base):
    metric_types = ['donor_topk_all', 'donor_topk', 'donor_auprc', 'donor_accuracy', 'donor_precision', 
                    'donor_recall', 'donor_f1', 'acceptor_topk_all', 'acceptor_topk', 'acceptor_auprc', 
                    'acceptor_accuracy', 'acceptor_precision', 'acceptor_recall', 'acceptor_f1', 
                    'accuracy', 'loss_batch', 'loss_every_update', 'learning_rate_every_epoch', 'learning_rate_every_batch']
    return {metric: f'{log_output_base}/{metric}.txt' for metric in metric_types}


def setup_device():
    device_str = "cuda" if torch.cuda.is_available() else "mps" if platform.system() == "Darwin" else "cpu"
    return torch.device(device_str)


def initialize_paths_inner(output_dir, project_name, flanking_size, exp_num, sequence_length, loss_fun, random_seed):
    MODEL_VERSION = f"SpliceAI_{project_name}_{flanking_size}_{exp_num}_rs{random_seed}"
    model_train_outdir = f"{output_dir}/{MODEL_VERSION}/{exp_num}/"
    model_output_base = f"{model_train_outdir}models/"
    log_output_base = f"{model_train_outdir}LOG/"
    log_output_train_base = f"{log_output_base}TRAIN/"
    log_output_val_base = f"{log_output_base}VAL/"
    log_output_test_base = f"{log_output_base}TEST/"
    for path in [model_output_base, log_output_train_base, log_output_val_base, log_output_test_base]:
        os.makedirs(path, exist_ok=True)
    return model_output_base, log_output_train_base, log_output_val_base, log_output_test_base


def load_data_from_shard(h5f, shard_idx, device, batch_size, params, shuffle=False):
    X = h5f[f'X{shard_idx}'][:].transpose(0, 2, 1)
    Y = h5f[f'Y{shard_idx}'][0, ...].transpose(0, 2, 1)
    X = torch.tensor(X, dtype=torch.float32)
    Y = torch.tensor(Y, dtype=torch.float32)
    ds = TensorDataset(X, Y)
    return DataLoader(ds, batch_size=batch_size, shuffle=shuffle, drop_last=True, pin_memory=True)


def classwise_accuracy(true_classes, predicted_classes, num_classes):
    class_accuracies = []
    for i in range(num_classes):
        true_positives = np.sum((predicted_classes == i) & (true_classes == i))
        total_class_samples = np.sum(true_classes == i)
        if total_class_samples > 0:
            accuracy = true_positives / total_class_samples
        else:
            accuracy = 0.0
        class_accuracies.append(accuracy)
    return class_accuracies


def metrics(batch_ypred, batch_ylabel, metric_files, run_mode):
    _, predicted_classes = torch.max(batch_ypred, 1)
    true_classes = torch.argmax(batch_ylabel, dim=1)
    true_classes = true_classes.numpy()
    predicted_classes = predicted_classes.numpy()
    true_classes_flat = true_classes.flatten()
    predicted_classes_flat = predicted_classes.flatten()
    accuracy = accuracy_score(true_classes_flat, predicted_classes_flat)
    precision, recall, f1, _ = precision_recall_fscore_support(true_classes_flat, predicted_classes_flat, average=None)
    class_accuracies = classwise_accuracy(true_classes, predicted_classes, 3)
    overall_accuracy = np.mean(class_accuracies)
    print(f"Overall Accuracy: {overall_accuracy}")
    for k, v in metric_files.items():
        with open(v, 'a') as f:
            if k == "accuracy":
                f.write(f"{overall_accuracy}\n")
    ss_types = ["Non-splice", "acceptor", "donor"]
    for i, (acc, prec, rec, f1_score) in enumerate(zip(class_accuracies, precision, recall, f1)):
        print(f"Class {ss_types[i]}\t: Accuracy={acc}, Precision={prec}, Recall={rec}, F1={f1_score}")
        if ss_types[i] == "Non-splice":
            continue
        for k, v in metric_files.items():
            with open(v, 'a') as f:
                if k == f"{ss_types[i]}_precision":
                    f.write(f"{prec}\n")
                elif k == f"{ss_types[i]}_recall":
                    f.write(f"{rec}\n")
                elif k == f"{ss_types[i]}_f1":
                    f.write(f"{f1_score}\n")
                elif k == f"{ss_types[i]}_accuracy":
                    f.write(f"{acc}\n")


def model_evaluation(batch_ylabel, batch_ypred, metric_files, run_mode, criterion):
    batch_ylabel = torch.cat(batch_ylabel, dim=0)
    batch_ypred = torch.cat(batch_ypred, dim=0)
    is_expr = (batch_ylabel.sum(axis=(1,2)) >= 1).cpu().numpy()
    if np.any(is_expr):
        subset_size = 1000
        indices = np.arange(batch_ylabel[is_expr].shape[0])
        subset_indices = np.random.choice(indices, size=min(subset_size, len(indices)), replace=False)
        batch_ylabel = batch_ylabel[is_expr][subset_indices, :, :]
        batch_ypred = batch_ypred[is_expr][subset_indices, :, :]
        Y_true_1 = batch_ylabel[:, 1, :].flatten().cpu().detach().numpy()
        Y_true_2 = batch_ylabel[:, 2, :].flatten().cpu().detach().numpy()
        Y_pred_1 = batch_ypred[:, 1, :].flatten().cpu().detach().numpy()
        Y_pred_2 = batch_ypred[:, 2, :].flatten().cpu().detach().numpy()
        acceptor_topk_accuracy, acceptor_auprc = print_topl_statistics(np.asarray(Y_true_1),
                            np.asarray(Y_pred_1), metric_files["acceptor_topk_all"], ss_type='acceptor', print_top_k=True)
        donor_topk_accuracy, donor_auprc = print_topl_statistics(np.asarray(Y_true_2),
                            np.asarray(Y_pred_2), metric_files["donor_topk_all"], ss_type='donor', print_top_k=True)
        if criterion == "cross_entropy_loss":
            loss = categorical_crossentropy_2d(batch_ylabel, batch_ypred)
        elif criterion == "focal_loss":
            loss = focal_loss(batch_ylabel, batch_ypred)
        for k, v in metric_files.items():
            with open(v, 'a') as f:
                if k == "loss_batch":
                    f.write(f"{loss.item()}\n")
                elif k == "donor_topk":
                    f.write(f"{donor_topk_accuracy}\n")
                elif k == "donor_auprc":
                    f.write(f"{donor_auprc}\n")
                elif k == "acceptor_topk":
                    f.write(f"{acceptor_topk_accuracy}\n")
                elif k == "acceptor_auprc":
                    f.write(f"{acceptor_auprc}\n")
        print("***************************************\n")
        metrics(batch_ypred, batch_ylabel, metric_files, run_mode)
    batch_ylabel = []
    batch_ypred = []
    return loss


def valid_epoch(model, h5f, idxs, batch_size, criterion, device, params, metric_files, flanking_size, run_mode):
    print(f"\033[1m{run_mode.capitalize()}ing model...\033[0m")
    model.eval()
    running_loss = 0.0
    np.random.seed(params["RANDOM_SEED"])
    shuffled_idxs = np.random.choice(idxs, size=len(idxs), replace=False)
    print("shuffled_idxs: ", shuffled_idxs)
    batch_ylabel = []
    batch_ypred = []
    print_dict = {}
    batch_idx = 0
    for i, shard_idx in enumerate(shuffled_idxs, 1):
        print(f"Shard {i}/{len(shuffled_idxs)}")
        loader = load_data_from_shard(h5f, shard_idx, device, batch_size, params, shuffle=False)
        pbar = tqdm(loader, leave=False, total=len(loader), desc=f'Shard {i}/{len(shuffled_idxs)}')
        for batch in pbar:
            DNAs, labels = batch[0].to(device), batch[1].to(device)
            DNAs, labels = clip_datapoints(DNAs, labels, params["CL"], CL_max, params["N_GPUS"])
            DNAs, labels = DNAs.to(torch.float32).to(device), labels.to(torch.float32).to(device)
            yp = model(DNAs)
            if criterion == "cross_entropy_loss":
                loss = categorical_crossentropy_2d(labels, yp)
            elif criterion == "focal_loss":
                loss = focal_loss(labels, yp)
            with open(metric_files["loss_every_update"], 'a') as f:
                f.write(f"{loss.item()}\n")
            running_loss += loss.item()
            batch_ylabel.append(labels.detach().cpu())
            batch_ypred.append(yp.detach().cpu())
            print_dict["loss"] = loss.item()
            pbar.set_postfix(print_dict)
            pbar.update(1)
            batch_idx += 1
        pbar.close()
    eval_loss = model_evaluation(batch_ylabel, batch_ypred, metric_files, run_mode, criterion)
    return eval_loss


def train_epoch(model, h5f, idxs, batch_size, criterion, optimizer, scheduler, device, params, metric_files, flanking_size, run_mode, global_batch_idx):
    print(f"\033[1m{run_mode.capitalize()}ing model...\033[0m")
    model.train()
    running_loss = 0.0
    np.random.seed(params["RANDOM_SEED"])
    shuffled_idxs = np.random.choice(idxs, size=len(idxs), replace=False)
    print("shuffled_idxs: ", shuffled_idxs)
    batch_ylabel = []
    batch_ypred = []
    print_dict = {}

    # Calculate total number of batches in the epoch
    total_batches_in_epoch = 0
    for shard_idx in idxs:
        # Load the loader to get its length
        loader = load_data_from_shard(h5f, shard_idx, device, batch_size, params, shuffle=False)
        total_batches_in_epoch += len(loader)

    for i, shard_idx in enumerate(shuffled_idxs, 1):
        print(f"Shard {i}/{len(shuffled_idxs)}")
        loader = load_data_from_shard(h5f, shard_idx, device, batch_size, params, shuffle=True)
        pbar = tqdm(loader, leave=False, total=len(loader), desc=f'Shard {i}/{len(shuffled_idxs)}')
        for batch in pbar:
            DNAs, labels = batch[0].to(device), batch[1].to(device)
            DNAs, labels = clip_datapoints(DNAs, labels, params["CL"], CL_max, params["N_GPUS"])
            DNAs, labels = DNAs.to(torch.float32).to(device), labels.to(torch.float32).to(device)
            optimizer.zero_grad()
            yp = model(DNAs)
            if criterion == "cross_entropy_loss":
                loss = categorical_crossentropy_2d(labels, yp)
            elif criterion == "focal_loss":
                loss = focal_loss(labels, yp)
            with open(metric_files["loss_every_update"], 'a') as f:
                f.write(f"{loss.item()}\n")
            loss.backward()
            optimizer.step()
            running_loss += loss.item()
            batch_ylabel.append(labels.detach().cpu())
            batch_ypred.append(yp.detach().cpu())
            print_dict["loss"] = loss.item()
            pbar.set_postfix(print_dict)
            pbar.update(1)

            # Update the scheduler
            epoch_fraction = global_batch_idx / total_batches_in_epoch
            scheduler.step(epoch_fraction)
            # Log current learning rate
            current_lr = scheduler.get_last_lr()[0]
            print_dict["lr"] = f"{current_lr:.6e}"        
            global_batch_idx += 1  # Increment global batch index
            with open(metric_files['learning_rate_every_batch'], 'a') as f:
                f.write(f"{current_lr}\n")
        pbar.close()
    eval_loss = model_evaluation(batch_ylabel, batch_ypred, metric_files, run_mode, criterion)
    return eval_loss, global_batch_idx


def calculate_metrics(y_true, y_pred):
    """Calculate metrics including precision, recall, f1-score, and accuracy."""
    precision, recall, f1, _ = precision_recall_fscore_support(y_true, y_pred, average='binary', zero_division=0)
    accuracy = accuracy_score(y_true, y_pred)
    return precision, recall, f1, accuracy


def threshold_predictions(y_probs, threshold=0.5):
    """Threshold probabilities to get binary predictions."""
    return (y_probs > threshold).astype(int)


def clip_datapoints(X, Y, CL, CL_max, N_GPUS):
    """
    Clip the input data points to the desired length.
    """
    rem = X.shape[0]%N_GPUS
    clip = (CL_max-CL)//2
    if rem != 0 and clip != 0:
        return X[:-rem, :, clip:-clip], Y[:-rem]
    elif rem == 0 and clip != 0:
        return X[:, :, clip:-clip], Y
    elif rem != 0 and clip == 0:
        return X[:-rem], Y[:-rem]
    else:
        return X, Y


def print_topl_statistics(y_true, y_pred, file, ss_type='acceptor', print_top_k=False):
    """
    Print top-k statistics for the given splice site type.
    """
    idx_true = np.nonzero(y_true == 1)[0]
    argsorted_y_pred = np.argsort(y_pred)
    sorted_y_pred = np.sort(y_pred)
    topkl_accuracy = []
    threshold = []
    for top_length in [0.5, 1, 2, 4]:
        num_elements = int(top_length * len(idx_true))
        if num_elements > len(y_pred):  # Check to prevent out-of-bounds access
            print(f"Warning: Requested top_length {top_length} with {len(idx_true)} true elements exceeds y_pred size of {len(y_pred)}. Adjusting to fit.")
            num_elements = len(y_pred)  # Adjust num_elements to prevent out-of-bounds error
        idx_pred = argsorted_y_pred[-int(top_length*len(idx_true)):]
        topkl_accuracy += [np.size(np.intersect1d(idx_true, idx_pred)) \
                  / float(min(len(idx_pred), len(idx_true))+1e-10)]
        threshold += [sorted_y_pred[-num_elements]]
    auprc = average_precision_score(y_true, y_pred)
    if print_top_k:
        print(f"\n\033[1m{ss_type}:\033[0m")
        print((("%.4f\t\033[91m%.4f\t\033[0m%.4f\t%.4f\t\033[94m%.4f\t\033[0m"
            + "%.4f\t%.4f\t%.4f\t%.4f\t%d") % (topkl_accuracy[0], topkl_accuracy[1], topkl_accuracy[2],
            topkl_accuracy[3], auprc, threshold[0], threshold[1],
            threshold[2], threshold[3], len(idx_true))))
    with open(file, 'a') as f:
        f.write((("%.4f\t%.4f\t%.4f\t%.4f\t%.4f\t"
          + "%.4f\t%.4f\t%.4f\t%.4f\t%d\n") % (topkl_accuracy[0], topkl_accuracy[1], topkl_accuracy[2],
          topkl_accuracy[3], auprc, threshold[0], threshold[1],
          threshold[2], threshold[3], len(idx_true))))
    return topkl_accuracy[1], auprc


def weighted_binary_cross_entropy(output, target, weights=None):    
    if weights is not None:
        assert len(weights) == 2
        loss = weights[1] * (target * torch.log(output+1e-10)) + \
               weights[0] * ((1 - target) * torch.log(1 - output+1e-10))
    else:
        loss = target * torch.log(output+1e-10) + (1 - target) * torch.log(1 - output+1e-10)
    return torch.neg(torch.mean(loss))


def categorical_crossentropy_2d(y_true, y_pred):
    """
    Compute 2D categorical cross-entropy loss.
    """
    return - torch.mean(y_true[:, 0, :]*torch.log(y_pred[:, 0, :]+1e-10)
                        + y_true[:, 1, :]*torch.log(y_pred[:, 1, :]+1e-10)
                        + y_true[:, 2, :]*torch.log(y_pred[:, 2, :]+1e-10))


def focal_loss(y_true, y_pred, alpha=0.25, gamma=2.0):
    """
    Compute 2D focal loss.
    """
    # Ensuring numerical stability
    gamma = 2
    epsilon = 1e-10
    return - torch.mean(y_true[:, 0, :]*torch.log(y_pred[:, 0, :]+epsilon) * torch.pow(torch.sub(1, y_pred[:, 0, :]), gamma)
                        + y_true[:, 1, :]*torch.log(y_pred[:, 1, :]+epsilon) * torch.pow(torch.sub(1, y_pred[:, 1, :]), gamma)
                        + y_true[:, 2, :]*torch.log(y_pred[:, 2, :]+epsilon) * torch.pow(torch.sub(1, y_pred[:, 2, :]), gamma))


def train_model(model, optimizer, scheduler, train_h5f, valid_h5f, test_h5f, train_idxs, val_idxs, test_idxs,
                model_output_base, args, device, params, 
                train_metric_files, valid_metric_files, test_metric_files):
    print(f"train_idxs (count: {len(train_idxs)}): ", train_idxs)
    print(f"val_idxs (count: {len(val_idxs)}): ", val_idxs)
    print(f"test_idxs (count: {len(test_idxs)}): ", test_idxs)
    best_val_loss = float('inf')
    epochs_no_improve = 0
    global_batch_idx = 0  # Initialize before the training loop
    for epoch in range(args.epochs):
        print(f"\n{'='*60}")
        # current_lr = optimizer.param_groups[0]['lr']
        # print(f">> Epoch {epoch + 1}; Current Learning Rate: {current_lr}")
        start_time = time.time()
        train_loss, global_batch_idx= train_epoch(model, train_h5f,
                        train_idxs, params["BATCH_SIZE"], args.loss, optimizer, scheduler, device, params, train_metric_files, args.flanking_size, run_mode="train", global_batch_idx=global_batch_idx)
        val_loss = valid_epoch(model, valid_h5f, val_idxs, params["BATCH_SIZE"], args.loss, device, 
                               params, valid_metric_files, args.flanking_size, "validation")
        test_loss = valid_epoch(model, test_h5f, test_idxs, params["BATCH_SIZE"], args.loss, device, 
                                params, test_metric_files, args.flanking_size, "test")
        print(f"Training Loss: {train_loss}")
        print(f"Validation Loss: {val_loss}")
        print(f"Testing Loss: {test_loss}")
        torch.save(model.state_dict(), f"{model_output_base}/model_{epoch}.pt")
        if args.early_stopping:
            if val_loss.item() < best_val_loss:
                best_val_loss = val_loss.item()
                torch.save(model.state_dict(), f"{model_output_base}/model_best.pt")
                print("New best model saved.")
                epochs_no_improve = 0
            else:
                epochs_no_improve += 1
                print(f"No improvement in validation loss for {epochs_no_improve} epochs.")
                if epochs_no_improve >= args.patience:
                    print("Early stopping triggered.")
                    break
        else:
            if val_loss.item() < best_val_loss:
                best_val_loss = val_loss.item()
                torch.save(model.state_dict(), f"{model_output_base}/model_best.pt")
                print("New best model saved.")
        current_lr = scheduler.get_last_lr()[0]
        with open(train_metric_files['learning_rate_every_epoch'], 'a') as f:
            f.write(f"{current_lr}\n")
        print(f">> Epoch {epoch + 1}; Final Learning Rate: {current_lr}")
        print(f"--- {time.time() - start_time:.2f} seconds ---")
        print("="*60)


"""
Filename: temperature_scaling.py
Author: Kuan-Hao Chao
Date: 2025-03-20
Description: Temperature scaling functions and classes for model calibration.
"""

import torch
from torch import nn
import torch.nn.functional as F
import numpy as np
from openspliceai.train_base.utils import *
from torch.utils.data import TensorDataset, DataLoader


def load_data_from_shard(h5f, shard_idx):
    X = h5f[f'X{shard_idx}'][:].transpose(0, 2, 1)
    Y = h5f[f'Y{shard_idx}'][0, ...].transpose(0, 2, 1)
    return X, Y


def get_validation_loader(h5f, idxs, batch_size):
    """
    Create a DataLoader for the validation data.
    """
    X_list, Y_list = [], []
    for shard_idx in idxs:
        X, Y = load_data_from_shard(h5f, shard_idx)
        X_list.append(X)
        Y_list.append(Y)
    X = np.concatenate(X_list, axis=0)
    Y = np.concatenate(Y_list, axis=0)
    X = torch.tensor(X, dtype=torch.float32)
    Y = torch.tensor(Y, dtype=torch.float32)
    print("X:", X.shape)
    print("Y:", Y.shape)
    dataset = TensorDataset(X, Y)
    loader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
    return loader


class ModelWithTemperature(nn.Module):
    """
    Wraps a model with class-wise temperature scaling for better calibration.
    """
    def __init__(self, model, num_classes):
        super().__init__()
        self.model = model
        # Each class gets its own temperature parameter
        self.temperature = nn.Parameter(torch.ones(num_classes) * 1.0)

    def forward(self, input):
        logits = self.model(input)
        return self.temperature_scale(logits)

    def temperature_scale(self, logits):
        """
        Class-wise temperature scaling: each logit for class c is divided by
        self.temperature[c]. If logits is [N, C], we broadcast over the batch dimension.
        """
        temperature = torch.clamp(self.temperature, min=0.05, max=5.0)
        return logits / temperature

    def save_temperature(self, filepath):
        """
        Save the temperature vector to a file.
        """
        torch.save(self.temperature.detach().cpu(), filepath)
        print(f"Temperature vector saved to {filepath}")

    def load_temperature(self, filepath, valid_loader, params):
        """
        Load the temperature parameter (vector) from a file, then compute and store
        logits/labels on the validation loader.
        """
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        temperature = torch.load(filepath, map_location=device)
        self.temperature = nn.Parameter(temperature.to(device))
        self.temperature.data = torch.clamp(self.temperature.data, min=0.05, max=5.0)
        print(f"Loaded temperature vector: {self.temperature.data.cpu().numpy()}")

        # Collect logits and labels
        logits_list, labels_list = [], []
        with torch.no_grad():
            for input, label in valid_loader:
                input, label = input.to(device), label.to(device)
                input, label = clip_datapoints(input, label, params["CL"], CL_max, params["N_GPUS"])
                logits = self.model(input)
                logits_list.append(logits.detach().cpu())
                labels_list.append(label.detach().cpu())

        logits = torch.cat(logits_list).permute(0, 2, 1).contiguous()
        labels = torch.cat(labels_list).permute(0, 2, 1).contiguous().argmax(dim=-1)

        # Flatten
        N, L, C = logits.shape
        logits = logits.view(-1, C).to(device)
        labels = labels.view(-1).long().to(device)

        self.logits, self.labels = logits, labels

    def set_temperature(self, valid_loader, params):
        """
        Tune the vector of temperature parameters using the validation set.
        """
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.to(device)
        self.model.eval()

        # Collect logits and labels
        logits_list, labels_list = [], []
        with torch.no_grad():
            for inputs, labels in valid_loader:
                inputs, labels = inputs.to(device), labels.to(device)
                inputs, labels = clip_datapoints(inputs, labels, params["CL"], CL_max, params["N_GPUS"])
                logits = self.model(inputs)
                logits_list.append(logits)
                labels_list.append(labels)

        logits = torch.cat(logits_list).permute(0, 2, 1).contiguous()
        labels = torch.cat(labels_list).permute(0, 2, 1).contiguous().argmax(dim=-1)

        # Flatten
        N, L, C = logits.shape
        logits = logits.view(-1, C)
        labels = labels.view(-1).long()

        self.logits, self.labels = logits.to(device), labels.to(device)

        # Define losses
        nll_criterion = nn.CrossEntropyLoss().to(device)
        ece_criterion = _ECELoss().to(device)

        # Print metrics before calibration
        self._compute_and_log_metrics(nll_criterion, ece_criterion, 'Before temperature scaling')

        # Optimize the temperature vector
        optimizer = torch.optim.Adam([self.temperature], lr=0.01)
        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
            optimizer, mode='min', factor=0.1, patience=2, verbose=True)

        best_loss = float('inf')
        best_temp = self.temperature.data.clone()
        patience_counter = 0
        max_epochs = 2000
        min_delta = 1e-6
        patience = 2

        for epoch in range(max_epochs):
            optimizer.zero_grad()
            loss = nll_criterion(self.temperature_scale(self.logits), self.labels)
            ece_loss = ece_criterion(self.temperature_scale(self.logits), self.labels)
            loss.backward()
            optimizer.step()
            self.temperature.data = torch.clamp(self.temperature.data, min=0.05, max=5.0)
            ece_loss = ece_loss.item()
            current_loss = loss.item()
            scheduler.step(current_loss)
            # Early stopping logic
            if best_loss - current_loss > min_delta:
                best_loss = current_loss
                best_temp = self.temperature.data.clone()
                patience_counter = 0
            else:
                patience_counter += 1

            print(f"Epoch {epoch+1}/{max_epochs}, Loss: {current_loss:.6f}, ECE: {ece_loss:.6f}",             
                  f"Temperature: {self.temperature.data.cpu().numpy()}")
            
            if patience_counter >= patience:
                print("Early stopping due to no improvement in loss.")
                break

        # Restore best temperature
        self.temperature.data = best_temp.to(device)
        print(f"Optimized temperature vector: {self.temperature.data.cpu().numpy()}")

        # Print metrics after calibration
        self._compute_and_log_metrics(nll_criterion, ece_criterion, 'After temperature scaling')

    def _compute_and_log_metrics(self, nll_criterion, ece_criterion, phase):
        logits_scaled = self.temperature_scale(self.logits)
        nll = nll_criterion(logits_scaled, self.labels).item()
        ece = ece_criterion(logits_scaled, self.labels).item()
        print(f'{phase} - NLL: {nll:.4f}, ECE: {ece:.4f}')

    def compute_ece_nll(self, logits, labels):
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        nll_criterion = nn.CrossEntropyLoss().to(device)
        ece_criterion = _ECELoss().to(device)
        nll = nll_criterion(logits.to(device), labels.to(device)).item()
        ece = ece_criterion(logits.to(device), labels.to(device)).item()
        return nll, ece


class _ECELoss(nn.Module):
    """
    Expected Calibration Error (ECE) Loss.
    """
    def __init__(self, n_bins=15):
        super().__init__()
        bin_boundaries = torch.linspace(0, 1, n_bins + 1)
        self.bin_lowers = bin_boundaries[:-1]
        self.bin_uppers = bin_boundaries[1:]

    def forward(self, logits, labels):
        softmaxes = F.softmax(logits, dim=1)
        confidences, predictions = torch.max(softmaxes, dim=1)
        accuracies = predictions.eq(labels)

        ece = torch.zeros(1, device=logits.device)
        for bin_lower, bin_upper in zip(self.bin_lowers, self.bin_uppers):
            in_bin = (confidences > bin_lower) & (confidences <= bin_upper)
            prop_in_bin = in_bin.float().mean()
            if prop_in_bin.item() > 0:
                accuracy_in_bin = accuracies[in_bin].float().mean()
                avg_confidence_in_bin = confidences[in_bin].mean()
                ece += torch.abs(avg_confidence_in_bin - accuracy_in_bin) * prop_in_bin
        return ece
